#include "Database.h"
#include "blowfish.h"

#include <SysEvtMgr.h> 	//	Needed for search for EvtSysEventAvail

static MemPtr  ApplicationInfoGetPtr(DmOpenRef);
static void InitAppInfoBlock(DmOpenRef, LocalID, LocalID);
static void ReplaceMemoPadHardKey(void);
static UInt32 SequentialSearch(UInt32 (* recordIDs)[2], UInt16, UInt32);


DmOpenRef dbCryptoPadRef;
DmOpenRef dbMemoPadRef;

// Added this flag to allow other places (notably Beam.c) to decide which
// MemoDB creator to target)

Boolean newMemoDB = true;

UInt16 	currentForm;
UInt16 	currentCategory;
FontID 	currentFieldFont;

FontID 	currentTableFont;
Boolean useSingleFont;
UInt16 	currentRecord;
UInt16 	topRecord;
Boolean saveBackup;
Boolean showLines;
UInt16 	editScrollPosition;
Boolean showAllCategories;
UInt8 	currentRotateOrder;
Boolean moveRecordOnTop;
IndexedColorType currentTableColor;
IndexedColorType currentFieldColor;
Boolean useLookupField;
UInt32	importantRecordID;
IndexedColorType blackColor;

Boolean wasEncrypted;
UInt16 	memosInCategory;
//UInt16 	tickAppButtonPushed = 0;
//Int16 	appButtonPushed = nullChr;
//Int16 	appButtonPushedModifiers = 0;
//Boolean	goneIntoImportantMemo = false;

Char 	categoryName[dmCategoryLength];
Char * 	currentPasswordP;
Char * 	globalPasswordP;
Char * 	prefsGlobalPasswordP;
privateRecordViewEnum PrivateRecordVisualStatus;
UInt32 	OSVersion;

/****************************************
 * OpenOrCreateMemoDatabase:
 * Open or creates the crypto database and
 * sets the default categories and
 * application preferences
 ****************************************
 */

DmOpenRef OpenMemoDatabase(UInt16 mode)
{
	DmOpenRef dbReference;

	/* Open the database */

    // First try to open the new "Memos" database, and fallback on the old
    // format if it can't be found.
	dbReference = DmOpenDatabaseByTypeCreator(memoPadType, memoPadCreatorNew, mode);
    if (! dbReference) {
        newMemoDB = false;
	    dbReference = DmOpenDatabaseByTypeCreator(memoPadType, memoPadCreator, mode);
    }
	ErrFatalDisplayIf(dbReference == 0, "Unable to open MemoPadDB !");

	return dbReference;
}

/****************************************
 * OpenOrCreateCryptoDatabase:
 * Open or creates the crypto database and
 * sets the default categories and
 * application preferences
 ****************************************
 */

DmOpenRef OpenOrCreateCryptoDatabase(UInt16 mode)
{
	Err err;
	DmOpenRef dbReference;

	LocalID dbID;
	LocalID appInfoID;

	UInt16 dbAttr;
	UInt16 dbAppVersion;
	UInt16 cardNo;

	/* Open the database */

	
	/* Test on which card the database is located
	 * as the new m5050/505 has extension SD slot.
	 */

	cardNo = GetCryptoPadCardNumber(&dbID);

	if (cardNo == dmErrCantFind) {
		cardNo = initCardNo;

	/* The database doesn't exist. Create it. */

		err = DmCreateDatabase(cardNo, cryptoPadName, cryptoPadCreator, cryptoPadType, false);
		ErrFatalDisplayIf(err, "Unable to create database");

		/* Open the database and get its reference */

		dbReference = DmOpenDatabaseByTypeCreator(cryptoPadType, cryptoPadCreator, mode);
		ErrFatalDisplayIf(dbReference == 0, "Could not find database.");

		/* Check the presence of the application info block */

		err = DmOpenDatabaseInfo(dbReference, &dbID, 0, 0, 0, 0);
		ErrFatalDisplayIf(err, "Unable to open database information");

		DmDatabaseInfo(cardNo, dbID, 0, &dbAttr, 0, 0,
							0, 0, 0, &appInfoID,
							0, 0, 0);
		
		dbAttr |= dmHdrAttrBackup;

		DmSetDatabaseInfo(cardNo, dbID, 0, &dbAttr, 0, 0,
						0, 0, 0, &appInfoID, 0, 0, 0);
		
		if (appInfoID == 0) {
			/* The application info block doesn't exist. Create it. */

			InitAppInfoBlock(dbReference, dbID, appInfoID);
		}

		/* Ask if user want to replace the MemoPad
		 * hard key to launch CryptoPad.
		 */

		 ReplaceMemoPadHardKey();

	} else {
		
		/* Open the database */

		dbReference = DmOpenDatabase (cardNo, dbID, mode);
	}

	DmOpenDatabaseInfo(dbReference, &dbID, 0, 0, 0, 0);
	
	DmDatabaseInfo(cardNo, dbID, 0, &dbAttr, &dbAppVersion, 0,
						0, 0, 0, &appInfoID, 0, 0, 0);
	
	if (dbAppVersion < dbGlobalPWVersion)
		InitAppInfoBlock(dbReference, dbID, appInfoID);

	/* Initialize variabled */

	currentPasswordP = 0;
	globalPasswordP = 0;

	return dbReference;
}


/*****************************************************
 * GetCryptoPadCardNumber:
 * Retrieve the card number on which
 * CryptoPadDB database was saved on.
 *****************************************************/

UInt16 GetCryptoPadCardNumber(LocalID * dbID)
{
	UInt16 cardNo;
	DmSearchStateType searchState;
	Err err;

	/* Test on which card the database is located
	 * as the new m500/505 has extension SD slot.
	 */

	err = DmGetNextDatabaseByTypeCreator (true, &searchState, cryptoPadType,
         		cryptoPadCreator, true, &cardNo, dbID);

    if (!err) return cardNo;

	return dmErrCantFind;
}

/*****************************************************
 * ReplaceMemoPadHardKey:
 * Replace the MemoPad
 * hard key to launch CryptoPad.
 *****************************************************/

static void ReplaceMemoPadHardKey(void)
{
	/* Ask if user want to replace the MemoPad
	 * hard key to launch CryptoPad.
	 */

	 if (PrefGetPreference(prefHard4CharAppCreator) != cryptoPadCreator)
		 if (FrmAlert(altLaunchCryptoPadWithHardKey) == 0)
			 PrefSetPreference(prefHard4CharAppCreator, cryptoPadCreator);
}

/***********************************************************************
 * InitAppInfoBlock:
 * Init the ApplicationInfo block.
 ***********************************************************************/
static void InitAppInfoBlock(DmOpenRef dbP, 
		LocalID dbID, LocalID appInfoID)
{
	UInt16 dbAppVersion;
	UInt16 cardNo;

	MemHandle handle;
	ApplicationInfoPtr appInfoP;
	ApplicationInfoPtr initAppInfoP;

	
	/* Create the new AppInof block if it does not exit
	* or it is a previous version.
	*/

	handle = DmNewHandle(dbP, sizeof(ApplicationInfoType));
	ErrFatalDisplayIf(handle == 0, "Unable to create application info block");

	/* Map the application info block to the database. */

	appInfoID = MemHandleToLocalID(handle);
	dbAppVersion = dbVersion;

	cardNo = GetCryptoPadCardNumber(&dbID);
	DmSetDatabaseInfo(cardNo, dbID, 0, 0, &dbAppVersion, 0,
		0, 0, 0, &appInfoID, 0, 0, 0);

	/* Clears the application info block */

	appInfoP = MemLocalIDToLockedPtr(appInfoID, cardNo);
	DmSet (appInfoP, 0, sizeof(ApplicationInfoType), 0);
	MemPtrUnlock(appInfoP);
}

/***********************************************************************
 * GetPrefsGlobalPassword:
 * Get the global password from
 * the ApplicationInfo.
 ***********************************************************************/
void GetPrefsGlobalPassword (void)
{
	Int8      length;
	ApplicationInfoPtr appInfoP;

	appInfoP = (ApplicationInfoPtr) ApplicationInfoGetPtr(dbCryptoPadRef); 

	length = StrLen(&(appInfoP->prefsGlobalPassword[0]));

	prefsGlobalPasswordP = MemPtrNew(length + 1);
	
	if (length != 0)
		StrCopy(prefsGlobalPasswordP, &(appInfoP->prefsGlobalPassword[0]));
	else prefsGlobalPasswordP = 0;
      
	MemPtrUnlock(appInfoP);
}

/***********************************************************************
 * SetPrefsGlobalPassword:
 * Get the global password from
 * the ApplicationInfo.
 ***********************************************************************/
void SetPrefsGlobalPassword(DmOpenRef dbP)
{
	ApplicationInfoPtr    appInfoP;
	ApplicationInfoPtr    appMemoInfoP;

	// Get a copy of the app info

	appInfoP = (ApplicationInfoPtr) ApplicationInfoGetPtr(dbP);
	ErrFatalDisplayIf(appInfoP == NULL, 
    	"Bad database (invalid or no app info block)");

	appMemoInfoP = (ApplicationInfoPtr) ApplicationInfoGetPtr(dbMemoPadRef);
	ErrFatalDisplayIf(appMemoInfoP == NULL, 
    	"Bad database (invalid or no app info block)");
      
	DmSet(appInfoP, 0, sizeof(ApplicationInfoType), 0);

	DmWrite(appInfoP, 0, appMemoInfoP, sizeof(ApplicationInfoType)
		- sizeof(appInfoP->prefsGlobalPassword));

	// Make the changes

	if (prefsGlobalPasswordP)
		DmWrite(appInfoP, sizeof(ApplicationInfoType)
			- sizeof(appInfoP->prefsGlobalPassword), prefsGlobalPasswordP,
			StrLen(prefsGlobalPasswordP));

	// Unlock app info

	MemPtrUnlock(appInfoP);
	MemPtrUnlock(appMemoInfoP);

}

/************************************************************
 * SetMemoPadSortOrder:
 * Set the sort order setting of MemoPad.
 *************************************************************/
Err SetMemoPadSortOrder(DmOpenRef dbP, Boolean sortOrder)
{
	MemoApplicationInfoPtr appInfoP;
	MemoApplicationInfoPtr	nilP = 0;


	appInfoP = (MemoApplicationInfoPtr) ApplicationInfoGetPtr (dbP);

	if (appInfoP->sortOrder != sortOrder) {
		DmWrite (appInfoP, (UInt32)&nilP->sortOrder, &sortOrder, sizeof(appInfoP->sortOrder));
		
		if (sortOrder == soAlphabetic) {
			DmInsertionSort(dbCryptoPadRef, 
				(DmComparF*) &CompareCryptoPadRecords, (Int16) sortOrder);
			DmInsertionSort(dbMemoPadRef, 
				(DmComparF*) &CompareMemoPadRecords, (Int16) sortOrder);
		}
	}

	MemPtrUnlock (appInfoP);	

	return 0;
}


/************************************************************
 * GetMemoPadSortOrder:
 * Get the sort order setting of MemoPad.
 *************************************************************/

UInt8 GetMemoPadSortOrder (DmOpenRef dbP)
{
	UInt8 sortOrder;
	MemoApplicationInfoPtr appInfoP;
		
	appInfoP = (MemoApplicationInfoPtr) ApplicationInfoGetPtr (dbP);

	sortOrder = appInfoP->sortOrder;

	MemPtrUnlock (appInfoP);	

	return (sortOrder);
}

/************************************************************
 * ApplicationInfoGetPtr:
 * Return a locked pointer to the ApplicationInfo or NULL
 *************************************************************/
static MemPtr   ApplicationInfoGetPtr(DmOpenRef dbP)
{
	UInt16   	  cardNum;
	LocalID    dbID;
	LocalID    appInfoID;
   
	if (DmOpenDatabaseInfo(dbP, &dbID, NULL, NULL, &cardNum, NULL))
    	return NULL;
	if (DmDatabaseInfo(cardNum, dbID, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &appInfoID, NULL, NULL, NULL))
	    return NULL;

	if (appInfoID == NULL)
		return NULL;
	else
    	return MemLocalIDToLockedPtr(appInfoID, cardNum);
   
}   

/**************************************
 * UpdateCryptoPadDB:
 * Update the records of CryptoPadDB
 * of the previous version to the new
 * CryptoPadRecordType.
 **************************************
 */

void UpdateCryptoPadDB(void)
{
	UInt16 recordNum = 0;
	UInt16 cardNo;
	MemHandle handle;
	LocalID dbID;

	while (true) {
		handle = DmQueryNextInCategory (dbCryptoPadRef, &recordNum, dmAllCategories);
		if (!handle) break;

		if (isEncrypted(recordNum, dbCryptoPadRef)) {
			MemHandle handle;
			CryptoRecordType newRecord;
			CryptoPackedRecordPtr packedP;
			Char * string;

			handle = DmGetRecord(dbCryptoPadRef, recordNum);
			packedP = MemHandleLock(handle);
	
		 	string = packedP->dataP;

			/* Create a new record with the data of the
			 * current one + standard values for the
			 * table and field font and color.
			 */

			newRecord.flag      = packedP->flag;
			newRecord.passwdLen =  packedP->passwdLen;
			newRecord.textLen   = packedP->textLen;
			newRecord.passwordP = string;
	
			string += newRecord.passwdLen + 1;
			newRecord.textP = string;
			newRecord.tableFont = stdFont;
			newRecord.fieldFont = stdFont;
			newRecord.tableColor = blackColor;
			newRecord.fieldColor = blackColor;

			MemHandleUnlock(handle);
			PackRecord(&newRecord, handle);

			DmReleaseRecord(dbCryptoPadRef, recordNum, true);
		}
		recordNum++;
	}

	/* Delete the CryptoFontDB database */

	cardNo = GetCryptoPadCardNumber(&dbID);
	dbID = DmFindDatabase (cardNo, cryptoFontName);

	if (dbID != 0) 
		if (FrmAlert(altDeleteCryptoFontDB) == 0)
			DmDeleteDatabase (cardNo, dbID);
}

/**************************************
 * SequentialSearch:
 * Routines uses to search for
 * the record number according to his uniqueID
 **************************************/

static UInt32 SequentialSearch(UInt32 (* recordIDs)[2], UInt16 numCells, UInt32 memoID)
{
#ifndef USES_SENTINEL
	UInt16 nIndex;

	for (nIndex = 0; nIndex < numCells; nIndex++) {
		if (recordIDs[nIndex][0] == memoID) return recordIDs[nIndex][1];
	}
	return 0;
#else
	UInt32* p;

	recordIDs[numCells][0] = memoID;
	for (p = &recordIDs[0][0]; p[0] != memoID; p += 2)
		/* nothing to do */;

	return ((p - &recordIDs[0][0]) < (numCells  * 2)) ? p[1]: 0;
#endif
}

/**************************************
 * SynchroniseRecord:
 * Synchronise the record
 * between both databases.
 **************************************
 */

void SynchroniseRecord(void)
{
	UInt16 numMemoPadRecord = 0;
	UInt16 recordNum;

	MemHandle handle;

	MemHandle memoHandle;
	MemoRecordType memoRecord;
	MemoRecordPtr memoRecordP;

	MemHandle strSynchroniseRecordH;
	Char* strSynchroniseRecordP;
	RectangleType tempRect;
	
	UInt32 (* recordIDs)[2];
	UInt16 numRecords;
	UInt16 nIndex;

	UInt32 memoID;
	UInt32 cryptoID;
	UInt16 recAttributes;

	/* Display a message to inform user of the
	 * MemoPad record synchronisation process
	 */

	strSynchroniseRecordH = DmGetResource('tSTR', strSynchroniseRecord);
	strSynchroniseRecordP = MemHandleLock(strSynchroniseRecordH);

	RctSetRectangle (&tempRect, 30, 80, 100, 20);
	WinEraseRectangle (&tempRect, 2);
	WinDrawRectangleFrame(popupFrame, &tempRect);


	FntSetFont(stdFont);
	DrawStringWithEllipsis (strSynchroniseRecordP, 15, 85, 130, justCenter);
	MemHandleUnlock(strSynchroniseRecordH);
	DmReleaseResource(strSynchroniseRecordH);

	numRecords = DmNumRecordsInCategory (dbCryptoPadRef, dmAllCategories);

#ifndef USES_SENTINEL
	recordIDs = MemPtrNew(numRecords * sizeof(UInt32) * 2);
#else
	recordIDs = MemPtrNew((numRecords + 1) * sizeof(UInt32) * 2);
#endif

	for (nIndex = 0; nIndex < numRecords; nIndex++) {
		recordIDs[nIndex][0] = 0;
		recordIDs[nIndex][1] = 0;
	}

	/* Checks if records have been deleted in the MemoPad database.
	 * During this first pass, we store both the CryptoPad and MemoPad
	 * record IDs in an array, to speed up subsequent lookups.
	 */

	nIndex = 0;
	recordNum = 0;

	while (true) {
		handle = DmQueryNextInCategory(dbCryptoPadRef, &recordNum, dmAllCategories);
		if (handle == 0) break;

		if (!isEncrypted(recordNum, dbCryptoPadRef)) {
			handle = DmQueryRecord(dbMemoPadRef, GetMemoPadRecordIndex(recordNum));

			if (handle == 0) {

				/* The record has been deleted from MemoPad.
				 * Remove it from CryptoPad.
				 */

				DmRemoveRecord(dbCryptoPadRef, recordNum);
				recordNum--;

			} else {

				/* The corresponding record was found in MemoPad. Extract
				 * both record IDs and synchronise the record categories
				 */

				DmRecordInfo(dbCryptoPadRef, recordNum, 0, &cryptoID, 0);
				DmRecordInfo(dbMemoPadRef, GetMemoPadRecordIndex(recordNum), &recAttributes, &memoID, 0);

				DmSetRecordInfo(dbCryptoPadRef, recordNum, &recAttributes, 0);

				/* Store both uniqueIDs in the array
				 */

				recordIDs[nIndex][0] = memoID;
				recordIDs[nIndex][1] = cryptoID;
				nIndex++;

				numMemoPadRecord++;
			}
		}
		recordNum++;
	}
	
	if (numMemoPadRecord  < DmNumRecords(dbMemoPadRef)) {

		recordNum = 0;

		while (true) {
			handle = DmQueryNextInCategory(dbMemoPadRef, &recordNum, dmAllCategories);
			if (handle == 0) break;

			/* Lookup corresponding CryptoPad record ID.
			 */

			DmRecordInfo(dbMemoPadRef, recordNum, &recAttributes, &memoID, 0);
			cryptoID = SequentialSearch(recordIDs, numRecords, memoID);
			
			if (cryptoID != 0) {
				recordNum++;
				continue;
			}

			/* Create a record in CryptoPadDB in relation with				
			 * the record from MemoPad.
			 */

			currentRecord = DmNumRecords(dbCryptoPadRef);

			/* Create a new record in the CryptoPad database */

			memoHandle = DmNewRecord(dbCryptoPadRef, &currentRecord, sizeof(MemoRecordType));
			if (memoHandle == 0) {
				FrmAlert(DeviceFullAlert);
				return;
			}

			memoRecordP = MemHandleLock(memoHandle);

			/* Write the default datas in the new record */

			memoRecord.flag = recDecrypted;
			memoRecord.uniqueID = memoID;

			if (useSingleFont == separateFont) {
				memoRecord.tableFont = stdFont;
				memoRecord.fieldFont = stdFont;
				memoRecord.tableColor = blackColor;
				memoRecord.fieldColor = blackColor;			
			} else {
				memoRecord.tableFont = currentTableFont;
				memoRecord.fieldFont = currentFieldFont;
				memoRecord.tableColor = currentTableColor;
				memoRecord.fieldColor = currentFieldColor;				
			}

			DmWrite(memoRecordP, 0, &memoRecord, sizeof(MemoRecordType));

			/* Unlock the block of the new record */

			MemPtrUnlock(memoRecordP);

			/* Set the category of the record */

			DmSetRecordInfo(dbCryptoPadRef, currentRecord, &recAttributes, 0);
			DmReleaseRecord(dbCryptoPadRef, currentRecord, true);

			recordNum++;
		}
	}

	WinEraseWindow ();

	/* Dispose of memory */

	MemPtrFree(recordIDs);
}

/**************************************
 * SetRecordFont:
 * Set the record table font
 * and field font according
 * to user choice.
 **************************************
 */
void SetRecordFont(UInt16 recordNum)
{
	MemHandle handle;

	handle = DmGetRecord(dbCryptoPadRef, recordNum);

	if (isEncrypted(recordNum, dbCryptoPadRef)) {
		CryptoPackedRecordPtr packedP;
		UInt16 nOffset;
	
		packedP = MemHandleLock(handle);

		/* Write in the packed record the new font settings */

		nOffset = sizeof(packedP->flag) +
				sizeof(packedP->passwdLen) +
				sizeof(packedP->textLen) +
				packedP->passwdLen + 1 +
				packedP->textLen + 1;

		DmSet(packedP, nOffset, sizeof(FontID), currentTableFont);
		nOffset += sizeof(FontID);

		DmSet(packedP, nOffset, sizeof(FontID), currentFieldFont);
		nOffset += sizeof(FontID);

		DmSet(packedP, nOffset, sizeof(IndexedColorType), currentTableColor);
		nOffset += sizeof(IndexedColorType);

		DmSet(packedP, nOffset, sizeof(IndexedColorType), currentFieldColor);

	} else {
		MemoRecordType memoRecord;
		MemoRecordPtr memoRecordP;

		memoRecordP = MemHandleLock(handle);
	
		/* Write the the new font settings in the record */

		memoRecord.flag = memoRecordP->flag;
		memoRecord.uniqueID = memoRecordP->uniqueID;
		memoRecord.tableFont = currentTableFont;
		memoRecord.fieldFont = currentFieldFont;
		memoRecord.tableColor = currentTableColor;
		memoRecord.fieldColor = currentFieldColor;

		DmWrite(memoRecordP, 0, &memoRecord, sizeof(MemoRecordType));

	}

	MemHandleUnlock(handle);
	DmReleaseRecord(dbCryptoPadRef, recordNum, true);
}

/**************************************
 * GetRecordFont:
 * Get the record table font
 * and field font according
 * to user choice.
 **************************************
 */
void GetRecordFont(UInt16 recordNum)
{
	MemHandle handle;

	handle = DmQueryRecord(dbCryptoPadRef, recordNum);

	if (isEncrypted(recordNum, dbCryptoPadRef) || wasEncrypted) {
		CryptoRecordType cryptoRecord;
		CryptoPackedRecordPtr packedP;

		/* Unpack the record */

		packedP = MemHandleLock(handle);
		UnpackRecord(&cryptoRecord, packedP);

		currentTableFont = cryptoRecord.tableFont;
		currentFieldFont = cryptoRecord.fieldFont;
		currentTableColor = cryptoRecord.tableColor;
		currentFieldColor = cryptoRecord.fieldColor;

	} else {
		MemoRecordPtr memoRecordP;

		memoRecordP = MemHandleLock(handle);

		currentTableFont = memoRecordP->tableFont;
		currentFieldFont = memoRecordP->fieldFont;
		currentTableColor = memoRecordP->tableColor;
		currentFieldColor = memoRecordP->fieldColor;
	}

	MemHandleUnlock(handle);
}

/*************************************************************
 * SetFieldColor:
 * Set the field color.
 ************************************************************/

 void SetFieldColor(IndexedColorType color)

 {
	/* Change the field color setting */

	if (OSVersion >= OS35) {
		RGBColorType rgb;

		WinIndexToRGB (color, &rgb);
		UIColorSetTableEntry (UIFieldText, &rgb);

	}
}

/****************************************
 * FontCharsInWidth:
 * Find if the first line of the textP fits
 * within the the strWidth. If it does not,
 * return text length in bytes that fits
 * and the exact width of this number of chars. 
 ****************************************
 */

Boolean FontCharsInWidth(Char * textP,
		UInt16* strWidth, UInt16* textLength)
{
	Char * 	ptr = StrChr (textP, linefeedChr);
	Boolean fits;
	
	*textLength = (ptr == NULL ? StrLen (textP) : (UInt16) (ptr - textP));
	
	FntCharsInWidth(textP, strWidth, textLength, &fits);

	return fits;
}

/**************************************
 * SynchroniseCategories:
 * Synchronise the categories between
 * MemoPad and CryptoPad.
 **************************************
 */

void SynchroniseCategories(void)
{
	UInt16 category;
	Char cryptoPadCategoryName[dmCategoryLength];
	Char memoPadCategoryName[dmCategoryLength];

	for (category = 1; category < dmRecNumCategories; category++) {

		CategoryGetName(dbCryptoPadRef, category, cryptoPadCategoryName);
		CategoryGetName(dbMemoPadRef, category, memoPadCategoryName);

		if ((StrCompare(cryptoPadCategoryName, memoPadCategoryName) != 0) &&
			(memoPadCategoryName[0] == '\0'))
				DmMoveCategory(dbCryptoPadRef, dmUnfiledCategory, category, true);
	}
}


/***************************************
 * ChangeCategory
 * This routine updates the global variables that keep track
 * of category information.  
 **************************************/

void ChangeCategory (UInt16 category)
{
	if (showAllCategories)
		memosInCategory = DmNumRecordsInCategory (dbCryptoPadRef, dmAllCategories);
	else memosInCategory = DmNumRecordsInCategory (dbCryptoPadRef, category);
		
	currentCategory = category;
	topRecord = 0;
}

/**************************************
 * RemoveCurrentPassword:
 * Clear the currentPasswordP variable.
 **************************************
 */

void RemoveCurrentPassword(void)
{
	MemSet(currentPasswordP, 0, 0);
	currentPasswordP = 0;
}

/**************************************
 * CreateRecord:
 * Creates an new empty memo.
 **************************************
 */

Boolean CreateRecord(void)
{
	MemHandle memoHandle;
	Char	zero = 0;
	UInt16	recAttributes = dmUnfiledCategory;

	/* Creates a new record at the end of the database */

	currentRecord = DmNumRecords (dbCryptoPadRef);

	/* Create a new record in the CryptoPad database */

	memoHandle = DmNewRecord(dbCryptoPadRef, &currentRecord, sizeof(MemoRecordType));
	if (!memoHandle) {
		FrmAlert(DeviceFullAlert);
		return (false);
	}

	/* Set the category of the record */

	recAttributes &= ~dmRecAttrCategoryMask;
	if (currentCategory != dmAllCategories) 
		recAttributes |= currentCategory;

	DmSetRecordInfo(dbCryptoPadRef, currentRecord, &recAttributes, 0);
	DmReleaseRecord(dbCryptoPadRef, currentRecord, true);


 	if (!ChangeRecordFromCryptoPadToMemoPad(&zero, currentRecord, 
		stdFont, stdFont, blackColor, blackColor))
			return (false);

	/* the category has gained one record */

	memosInCategory++;

	return (true);
}

/************************************
 * ChangeRecordFromMemoPadToCryptoPad:
 * Change the structure of the CryptoPad
 * record, encrypt it and delete the
 * record in MemoPad.
 ************************************
 */

Boolean ChangeRecordFromMemoPadToCryptoPad(char * textP, UInt16 cryptoRecordNum, 
	FontID tableFont, FontID fieldFont, IndexedColorType tableColor, IndexedColorType fieldColor)
{
	MemHandle handle;
	CryptoRecordType newRecord;

	UInt16 index;
	UInt16 textLength, passwordLength;

	/* Get the index in MemoPadDB of the record to encrypt */

	index = GetMemoPadRecordIndex(cryptoRecordNum);

	handle = DmGetRecord(dbCryptoPadRef, cryptoRecordNum);

	/* Current record now contains the record index.
	 * Sets an initial value for the record.
	 */
	
	passwordLength = StrLen(currentPasswordP);
	textLength = StrLen(textP);

	newRecord.flag      = recDecrypted;
	newRecord.passwdLen =  passwordLength;
	newRecord.textLen   = textLength;
	newRecord.passwordP = currentPasswordP;
	newRecord.textP = textP;
	if (useSingleFont == separateFont) {
		newRecord.tableFont = tableFont;
		newRecord.fieldFont = fieldFont;
		newRecord.tableColor = tableColor;
		newRecord.fieldColor = fieldColor;
	} else {
		newRecord.tableFont = stdFont;
		newRecord.fieldFont = stdFont;
		newRecord.tableColor = blackColor;
		newRecord.fieldColor = blackColor;
	}

	PackRecord(&newRecord, handle);

	DmReleaseRecord(dbCryptoPadRef, cryptoRecordNum, true);
	
	/* Delete the record from MemoPadDB */

	DmDeleteRecord (dbMemoPadRef, index);
	DmMoveRecord (dbMemoPadRef, index, DmNumRecords (dbMemoPadRef));

	/* Encrypt the record if the password is not empty */

	if (currentPasswordP != 0)
		EncryptRecord(cryptoRecordNum, currentPasswordP);
}

/************************************
 * ChangeRecordFromCryptoPadToMemoPad:
 * Create a new record in MemoPad with
 * the correct text and change the 
 * structure of the CryptoPad record.
 ************************************
 */

Boolean ChangeRecordFromCryptoPadToMemoPad(char * textP, UInt16 cryptoRecordNum, 
	FontID tableFont, FontID fieldFont, IndexedColorType tableColor, IndexedColorType fieldColor)
{
	MemPtr		p;
	MemHandle handle;
	MemoRecordType memoRecord;
	MemoRecordPtr memoRecordP;

	UInt16 index;	
	UInt16 recAttributes;
	UInt16 nLength;			
	UInt32 recordID;

	index = dmMaxRecordIndex;

	/* Create a new record in the MemoPad database */

	nLength = StrLen(textP);
	handle = DmNewRecord(dbMemoPadRef, &index, nLength + 1);
	if (!handle) {
		FrmAlert(DeviceFullAlert);
		return (false);
	}

	/* Lock down the block containing the new record */

	p = MemHandleLock(handle);

	/* Write the field's text to the database record */

	DmWrite(p, 0, textP, nLength + 1);

	/* Unlock the block of the new record */

	MemPtrUnlock(p);

	/* Set the category of the record */
	DmRecordInfo(dbCryptoPadRef, cryptoRecordNum, &recAttributes, 0, 0);

	DmReleaseRecord(dbMemoPadRef, index, true);

	DmSetRecordInfo(dbMemoPadRef, index, &recAttributes, 0);

	/* Change the structure of the record from CryptoPadDB */

	DmRecordInfo (dbMemoPadRef, index, 0, &recordID, 0);

	handle = DmGetRecord(dbCryptoPadRef, cryptoRecordNum);
	MemHandleResize(handle, sizeof(MemoRecordType));
	memoRecordP = MemHandleLock(handle);

	/* Write the default datas in the new record */

	memoRecord.flag = recDecrypted;
	memoRecord.uniqueID = recordID;
	if (useSingleFont == separateFont) {
		memoRecord.tableFont = tableFont;
		memoRecord.fieldFont = fieldFont;
		memoRecord.tableColor = tableColor;
		memoRecord.fieldColor = fieldColor;
	} else {
		memoRecord.tableFont = stdFont;
		memoRecord.fieldFont = stdFont;
		memoRecord.tableColor = blackColor;
		memoRecord.fieldColor = blackColor;
	}
	DmWrite(memoRecordP, 0, &memoRecord, sizeof(MemoRecordType));

	/* Unlock the block of the new record */

	MemHandleUnlock(handle);
	DmReleaseRecord(dbCryptoPadRef, cryptoRecordNum, true);

	return (true);
}

/************************************
 * CompareMemoPadRecords:
 * This function is used by the
 * sort algorith routine.
 ************************************
 */

Int16 CompareMemoPadRecords(MemPtr rec1, MemPtr rec2, Int16 sortOrder)
{
	if (sortOrder == soAlphabetic) {
		return StrCompare(rec1, rec2);
	} else return 0;
}


/************************************
 * CompareCryptoPadRecords:
 * This function is used by the
 * sort algorith routine.
 ************************************
 */

Int16 CompareCryptoPadRecords(PackedRecordPtr rec1, 
		PackedRecordPtr rec2, Int16 sortOrder)
{
	char * textP1;
	char * textP2;
	MemHandle handle1, handle2;
	Int8 strComparison;
	Boolean unlockRec1 = false, unlockRec2 = false;

	if (sortOrder == soAlphabetic) {
		/* store in textP1 the record text according to its status */

		if (rec1->flag == recEncrypted) {

			// If the record is encrypted...

			CryptoRecordType record1;

			UnpackRecord(&record1, (CryptoPackedRecordPtr) rec1);
			textP1 = record1.textP;

		} else {
			// If the record is not encrypted ...

			UInt16 index;

			if (DmFindRecordByID (dbMemoPadRef,((MemoRecordPtr) rec1)->uniqueID, &index)) return (0);

			handle1 = DmQueryRecord(dbMemoPadRef, index);
			textP1 = MemHandleLock(handle1);

			unlockRec1 = true;
		}

		/* store in textP1 the record text according to its status */

		if (rec2->flag == recEncrypted) {

			// If the record is encrypted...

			CryptoRecordType record2;
			UnpackRecord(&record2, (CryptoPackedRecordPtr) rec2);
			textP2 = record2.textP;

		} else {
			// If the record is not encrypted...

			UInt16 index;
		
			if (DmFindRecordByID (dbMemoPadRef,((MemoRecordPtr) rec2)->uniqueID, &index)) return (0);

			handle2 = DmQueryRecord(dbMemoPadRef, index);
			textP2 = MemHandleLock(handle2);

			unlockRec2 = true;
		}

		strComparison = StrCompare(textP1, textP2);

		/* if a record was not encrypted,
		 *  a handle is locked and we have to unlock it
		 */

		if (unlockRec1) MemHandleUnlock(handle1);
		if (unlockRec2) MemHandleUnlock(handle2);

		return strComparison;

	} else return 0;
}

/******************************************
 * isEncrypted:
 * Tells the status of the given record
 ******************************************
 */

Boolean isEncrypted(UInt16 recordNum, DmOpenRef cryptoDBP)
{
	Boolean bResult;
	MemHandle handle;
	PackedRecordPtr packedP;

	handle = DmQueryRecord(cryptoDBP, recordNum);
	if (!handle) return;

	packedP = MemHandleLock(handle);
	bResult = (packedP->flag == recEncrypted);
	MemHandleUnlock(handle);

	return bResult;
}

/******************************************
 * GetImportantState:
 * Return true if the memo is
 * marked 'Important'.
 ******************************************
 */

Boolean GetImportantState(void)
{
	UInt32 uniqueID;

	DmRecordInfo (dbCryptoPadRef, currentRecord, NULL, &uniqueID, NULL);

	if (uniqueID == importantRecordID) return true;
	else return false;
}

/**************************************
 * GetMemoPadRecordIndex:
 * Find the MemoPad record index
 * of the CryptoPad record at the
 * recordNum index.
 **************************************
 */

UInt16 GetMemoPadRecordIndex(UInt16 recordNum)
{
	MemHandle handle;
	MemoRecordPtr memoRecordP;
	UInt16 index;

	/* Get the uniqueID */	

	handle = DmQueryRecord(dbCryptoPadRef, recordNum);
	memoRecordP = MemHandleLock(handle);

	DmFindRecordByID (dbMemoPadRef,memoRecordP->uniqueID, &index);

	MemHandleUnlock(handle);

	return index;
}

/**************************************
 * UnpackRecord:
 * Unpacks the record to identify the
 * address of the different strings in
 * the record.
 **************************************
 */

void UnpackRecord(CryptoRecordPtr recordP, CryptoPackedRecordPtr packedRecordP)
{
	Char * string = packedRecordP->dataP;

	recordP->flag      = packedRecordP->flag;
	recordP->passwdLen = packedRecordP->passwdLen;
	recordP->textLen   = packedRecordP->textLen;

	recordP->passwordP = string;

	string += recordP->passwdLen + 1;

	recordP->textP = string;
	string += recordP->textLen + 1;

	MemMove(&recordP->tableFont, string, sizeof(FontID));
	string += sizeof(recordP->tableFont);

	MemMove(&recordP->fieldFont, string, sizeof(FontID));
	string += sizeof(recordP->fieldFont);

	MemMove(&recordP->tableColor, string, sizeof(IndexedColorType));
	string += sizeof(recordP->tableColor);

	MemMove(&recordP->fieldColor, string, sizeof(IndexedColorType));
	string += sizeof(recordP->fieldColor);
}

/**************************************
 * PackRecord:
 * Packs the record into one structure
 **************************************
 */

Err PackRecord(CryptoRecordPtr recordP, MemHandle record)
{
	UInt16 nLength;
	UInt16 offset;

	offset = 0;
	nLength = sizeof(recordP->flag) +
				sizeof(recordP->passwdLen) +
				sizeof(recordP->textLen) +
				recordP->passwdLen + 1 +
				recordP->textLen + 1 +
				sizeof(recordP->tableFont) +
				sizeof(recordP->fieldFont) +
				sizeof(recordP->tableColor) +
				sizeof(recordP->fieldColor);


	/* Resize the handle */

	if (MemHandleResize(record, nLength) == 0) {
		Char * ptr;

		ptr = MemHandleLock(record);
		offset = 0;

		DmSet(ptr, offset, 1, recordP->flag);
		offset += sizeof(recordP->flag);

		DmWrite(ptr, offset, &(recordP->passwdLen),
							sizeof(recordP->passwdLen));
		offset += sizeof(recordP->passwdLen);

		DmWrite(ptr, offset, &(recordP->textLen),
							sizeof(recordP->textLen));
		offset += sizeof(recordP->textLen);

		DmWrite(ptr, offset, recordP->passwordP, recordP->passwdLen + 1);
		offset += recordP->passwdLen + 1;

		DmWrite(ptr, offset, recordP->textP, recordP->textLen + 1);
		offset += recordP->textLen + 1;

		DmSet(ptr, offset, sizeof(recordP->tableFont), recordP->tableFont);
		offset += sizeof(recordP->tableFont);

		DmSet(ptr, offset, sizeof(recordP->fieldFont), recordP->fieldFont);
		offset += sizeof(recordP->fieldFont);

		DmSet(ptr, offset, sizeof(recordP->tableColor), recordP->tableColor);
		offset += sizeof(recordP->tableColor);

		DmSet(ptr, offset, sizeof(recordP->fieldColor), recordP->fieldColor);
		offset += sizeof(recordP->fieldColor);

		MemHandleUnlock(record);
		return 0;

	} else return 1;
}



/***************************************
 * Search
 * This routine searchs the memo database for records 
 * containing the string passed.
 ***************************************/
void Search (FindParamsPtr findParams, Boolean appIsActive)
{
	UInt16 pos;
	Char * header;
	MemHandle recordH, headerStringH;
	RectangleType r;
	Boolean done, match;
	
	DmOpenRef memoDBP, cryptoDBP;
	DmSearchStateType searchState;
	Boolean encrypted = false;
	Err err, err2;
	UInt16 cardNum;
	UInt16 recordNum;
	LocalID memoPadDbID, cryptoPadDbID;

	Char * textP;
	UInt16 index;

    // First try to open the new "Memos" database, and fallback on the old
    // format if it can't be found.
	err = DmGetNextDatabaseByTypeCreator (true, &searchState, memoPadType,
         		memoPadCreatorNew, true, &cardNum, &memoPadDbID);
    if (err == dmErrCantFind) {
	    err = DmGetNextDatabaseByTypeCreator (true, &searchState, memoPadType,
         		memoPadCreator, true, &cardNum, &memoPadDbID);
    }
	err2 = DmGetNextDatabaseByTypeCreator (true, &searchState, cryptoPadType,
         		cryptoPadCreator, true, &cardNum, &cryptoPadDbID);

     if (err || err2)
		{
		findParams->more = false;
		return;
		}

   // Open the MemoPad database.

   memoDBP = DmOpenDatabase(cardNum, memoPadDbID, findParams->dbAccesMode);

   // Open the CryptoPad database.

   cryptoDBP = DmOpenDatabase(cardNum, cryptoPadDbID, findParams->dbAccesMode);
   
   if ((!memoDBP) || (!cryptoDBP))
   	{
      findParams->more = false;
      return;
   	}

	// Display the heading line.

	headerStringH = DmGetResource('tSTR', strCryptoHeader);
	header = MemHandleLock(headerStringH);
	done = FindDrawHeader(findParams, header);
   	MemHandleUnlock(headerStringH);   
   	DmReleaseResource(headerStringH);   
   	if (done) 
      goto Exit;

	// Search the memos for the "find" string.

	recordNum = findParams->recordNum;
	while (true)
		{
		// Because applications can take a long time to finish a find when
		// the result may be on the screen or for other reasons, users like
		// to be able to stop the find.  Stop the find if an event is pending.
		// This stops if the user does something with the device.  Because
		// this call slows winDown the search we perform it every so many 
		// records instead of every record.  The response time should still
		// be short without introducing much extra work to the search.
		
		// Note that in the implementation below, if the next 16th record is  
		// secret the check doesn't happen.  Generally this shouldn't be a 
		// problem since if most of the records are secret then the search 
		// won't take long anyways!

		if ((recordNum & 0x000f) == 0 &&			// every 16th record
			EvtSysEventAvail(true))
			{
			// Stop the search process.
			findParams->more = true;
			break;
			}

		recordH = DmQueryNextInCategory (cryptoDBP, &recordNum, dmAllCategories);


	 	if (! recordH){
				findParams->more = false;			
				break;
		}

		match = false;

		if ((isEncrypted(recordNum, cryptoDBP)) 
			|| (appIsActive && (currentPasswordP != 0) && (recordNum == currentRecord))) {

			if (appIsActive)
				if (globalPasswordP) {
					CryptoPackedRecordPtr cryptoRecP;
					CryptoRecordType cryptoRecord;

					if (currentPasswordP == 0) {
						encrypted = true;
						DecryptRecord(recordNum, globalPasswordP);
					}

					cryptoRecP = MemHandleLock(recordH);
					UnpackRecord(&cryptoRecord, cryptoRecP);

					textP = MemPtrNew(StrLen(cryptoRecord.textP) + 1);
					StrCopy(textP, cryptoRecord.textP);

					match = FindStrInStr (cryptoRecord.textP, findParams->strToFind, &pos);

					MemHandleUnlock(recordH);

					if (currentPasswordP == 0)
						DmReleaseRecord(cryptoDBP, recordNum, false);
			}	
		} else {
			MemoRecordPtr memoRecordP;
			MemHandle handle;
			MemPtr memoRecP;

			encrypted = false;

			/* Get the uniqueID */	

			memoRecordP = MemHandleLock(recordH);
			DmFindRecordByID (memoDBP,memoRecordP->uniqueID, &index);
			MemHandleUnlock(recordH);

			handle = DmQueryRecord(memoDBP, index);
			memoRecP = MemHandleLock(handle);

			textP = MemPtrNew(StrLen(memoRecP) + 1);
			StrCopy(textP, memoRecP);

			match = FindStrInStr (textP, findParams->strToFind, &pos);
			MemHandleUnlock(handle);
		}

		if (match)
			{
			// Add the match to the find parameter block,  if there is no room to
			// display the match the following function will return true.

			done = FindSaveMatch (findParams, recordNum, pos, 0, 0, cardNum, cryptoPadDbID);

			if (!done)
				{
				// Get the bounds of the region where we will draw the results.

				FindGetLineBounds (findParams, &r);
	
				// Display the title of the description.	

				DrawStringWithEllipsis  (textP, r.topLeft.x+1, 
							r.topLeft.y, r.extent.x-2, justLeft);
				findParams->lineNumber++;
				}
			}

			if (encrypted && globalPasswordP)
				EncryptRecord(recordNum, globalPasswordP);

		if (done) break;

		recordNum++;
		}

Exit:
	MemPtrFree(textP);
	DmCloseDatabase (memoDBP);
	DmCloseDatabase (cryptoDBP);
}


/***************************************
 * GoToItem
 * This routine is an entry point of the memo application.
 * It is generally called as the result of hitting a 
 * "Go to" button in the text search dialog.  The record
 * identifies by the parameter block passed will be display,
 * with the character range specified highlighted.
 ***************************************/
void GoToItem (GoToParamsPtr goToParams, Boolean launchingApp)
{
	UInt16 recordNum = 0;
	UInt16 attr;
	UInt32 uniqueID;
	EventType event;

	LocalID memoPadDbID, cryptoPadDbID;
	MemHandle handle;

	DmOpenDatabaseInfo (dbCryptoPadRef, &cryptoPadDbID, 0, 0, 0, 0);
	DmOpenDatabaseInfo (dbMemoPadRef, &memoPadDbID, 0, 0, 0, 0);

	if (goToParams->dbID == memoPadDbID) {
		while (true) {
			handle = DmQueryNextInCategory(dbCryptoPadRef, &recordNum, dmAllCategories);
			if (handle == 0) break;

			if (!isEncrypted(recordNum, dbCryptoPadRef))
				if (GetMemoPadRecordIndex(recordNum) == goToParams->recordNum) break;
			recordNum++;
		}
	} else if (goToParams->dbID == cryptoPadDbID)
		recordNum = goToParams->recordNum;

	//   DmRecordInfo is side effect of that if statement

	if (DmRecordInfo (dbCryptoPadRef, recordNum, &attr, &uniqueID, NULL) != 0)
		{
		FrmGotoForm(frmMainForm);
		return;
		}
	// Make the item the first item displayed.

	topRecord = recordNum;

	// Change the current category if necessary.

	if (currentCategory != dmAllCategories)
		ChangeCategory (attr & dmRecAttrCategoryMask);

	// If the application is already running, close all open forms.  This 
	// may cause in the database record to be reordered, so we'll find the 
	// records index by its unique id.

	if (! launchingApp)
		{
		FrmCloseAllForms ();
		DmFindRecordByID (dbCryptoPadRef, uniqueID, &recordNum);
		}
		
	// Send an event to goto a form and select the matching text.

	MemSet (&event, sizeof(EventType), 0);

	event.eType = frmLoadEvent;
	event.data.frmLoad.formID = frmEditForm;
	EvtAddEventToQueue (&event);
 
	event.eType = frmGotoEvent;
	event.data.frmGoto.recordNum = recordNum;
	event.data.frmGoto.matchPos = goToParams->matchPos;
	event.data.frmGoto.matchLen = goToParams->searchStrLen;
	event.data.frmGoto.matchFieldNum = goToParams->matchFieldNum;
	event.data.frmGoto.formID = frmEditForm;
	EvtAddEventToQueue (&event);

	currentForm = frmEditForm;
}
