// $Header: /cvsroot/cryptopad/cryptopad_classic/v4/src/DetailsForm.c,v 1.3 2005/10/10 05:12:38 jbauman Exp $

#include "DetailsForm.h"
#include "Database.h"
#include "Utils.h"

static UInt16 DetailsFormApplyChanges(void) DETAILS_SECTION;
static void DetailsFormInit(void) DETAILS_SECTION;
static Boolean importantChecked;

UInt16 category;
Boolean categoryEdited;


static void DetailsFormInit(void)
{
	UInt16 recAttributes;
	ControlPtr controlP;

	DmRecordInfo(dbCryptoPadRef, currentRecord, &recAttributes, 0, 0);

	/* Updates the current record category */

	category = (recAttributes & dmRecAttrCategoryMask);

	CategoryGetName(dbMemoPadRef, category, categoryName);
	CategorySetTriggerLabel(GetActiveFormObjectPtr(popDetailsCategory), categoryName);

	/* Updates the private checkbox */

	controlP = GetActiveFormObjectPtr(chkDetailsPrivate);
	CtlSetValue(controlP, (recAttributes & dmRecAttrSecret) == dmRecAttrSecret);

	/* Updates the important checkbox */
/*
	controlP = GetActiveFormObjectPtr(chkDetailsImportant);
	importantChecked = (GetImportantState() == true);
	CtlSetValue(controlP, importantChecked);
*/
}

Boolean DetailsFormDeleteCurrentRecord(void)
{
	FormPtr formP;
	FieldPtr fieldP;

	Char *  textP;
	Boolean empty;
	UInt16 index;

	formP = FrmGetFormPtr(frmEditForm);
	fieldP = FrmGetObjectPtr(formP, FrmGetObjectIndex(formP, fldEditText));

	if (!wasEncrypted)
		index = GetMemoPadRecordIndex(currentRecord);

	// Find out if the field is empty.
	textP = FldGetTextPtr (fieldP);
	empty = (*textP == 0);

	// Display an alert to confirm the operation.
	if (!empty) {
		if (!wasEncrypted)	{
	
			UInt16 btn;
			FormPtr alert;
			Boolean backup;
			UInt16 ctlIndex;

			alert = FrmInitForm (frmDeleteForm);
	
			ctlIndex = FrmGetObjectIndex (alert, chkDeleteBackup);
			FrmSetControlValue (alert, ctlIndex, saveBackup);
			btn = FrmDoDialog (alert);
			backup = FrmGetControlValue (alert, ctlIndex);
	
			FrmDeleteForm (alert);

			if (btn  == btnDeleteCancel)
				return (false);
	
			saveBackup = backup;
		} else {
			if (FrmAlert(altConfirmDeleteAlert) == 1) return (false);
		}
	}

	// Clear the handle value in the field, otherwise the handle
	// will be freed when the form is disposed of.
	FldSetTextHandle (fieldP, 0);

	// Delete or archive the record.
	if (!wasEncrypted) {
		if (empty && (! FldDirty (fieldP)))
			DmRemoveRecord(dbMemoPadRef, index);
		else {
			if (saveBackup)
				DmArchiveRecord (dbMemoPadRef, index);
			else
				DmDeleteRecord (dbMemoPadRef, index);
			DmMoveRecord (dbMemoPadRef, index, DmNumRecords (dbMemoPadRef));
		}
	} 
	DmRemoveRecord (dbCryptoPadRef, currentRecord);

	memosInCategory--;
	currentRecord = noRecordSelected;
	wasEncrypted = recDecrypted;

	RemoveCurrentPassword();

	return (true);
}


static UInt16 DetailsFormApplyChanges(void)
{
	UInt16 recAttributes;
	Boolean private;
	ControlPtr controlP;
	UInt16 updateCode = 0;
	UInt32 recordID;
	Boolean dirty = false;

	DmRecordInfo(dbCryptoPadRef, currentRecord, &recAttributes, &recordID, 0);

	/* Updates the memo category based on the category list */
	if (currentCategory != category) {
		recAttributes &= ~dmRecAttrCategoryMask;
		recAttributes |= category;
		dirty = true;

		updateCode |= updateCategoryChanged;
	}

	/* If the current category was deleted, renamed, or merged with
	* another category, then the list view needs to be redrawn.
	*/
	if (categoryEdited)
		{
		currentCategory = category;
		updateCode |= updateCategoryChanged;
		}


	/* Updates the secret bit based on the private checkbox */

	controlP = GetActiveFormObjectPtr(chkDetailsPrivate);
	private = (CtlGetValue(controlP) != 0);

	if (private != ((recAttributes & dmRecAttrSecret) == dmRecAttrSecret)) {
		if (private)  {
			recAttributes |= dmRecAttrSecret;
			if (PrivateRecordVisualStatus == showPrivateRecords)
				FrmAlert(privateRecordInfoAlert);
		} else recAttributes &= ~dmRecAttrSecret;

		dirty = true;
	}

	/* Apply changes in the importantRecordID variable */
/*
	controlP = GetActiveFormObjectPtr(chkDetailsImportant);
	if (importantChecked != CtlGetValue(controlP)) {
		if (CtlGetValue(controlP))
			importantRecordID = recordID;
		else importantRecordID = noRecordSelectedID;

		updateCode |= updateImportantDisplay;
	}
*/
	/* Apply changes in the record attributes */

	if (dirty) {
		recAttributes |= dmRecAttrDirty;
		DmSetRecordInfo(dbCryptoPadRef, currentRecord, &recAttributes, 0);
		if (!wasEncrypted)
			DmSetRecordInfo(dbMemoPadRef, GetMemoPadRecordIndex(currentRecord), &recAttributes, 0);	
	}

	return (updateCode);
}


Boolean DetailsFormHandleEvent(EventPtr eventP)
{
	Boolean handled = false;
	Int16 updateCode;
	Boolean categoryUpdated;

	switch (eventP->eType) {
		case ctlSelectEvent:
			switch (eventP->data.ctlSelect.controlID) {
				case btnDetailsOK:
					updateCode = DetailsFormApplyChanges();
					FrmReturnToForm(currentForm);
	
					/* if the current category has be deleted, renamed or
					 * merged with another category, we need to update the
					 * edit form.
					 */
					if (updateCode)
						FrmUpdateForm (currentForm, updateCode);			
					handled = true;
					break;

				case btnDetailsCancel:
					if (categoryEdited)
						FrmUpdateForm (currentForm, updateCategoryChanged);
					FrmReturnToForm(currentForm);
					handled = true;
					break;

				/* Remove the current memo and return to the main form */
				case btnDetailsDelete:
					if (DetailsFormDeleteCurrentRecord()) {
						FormPtr		frm;

						frm = FrmGetActiveForm();
						FrmEraseForm (frm);
						FrmDeleteForm (frm);
						FrmCloseAllForms();
						FrmGotoForm(frmMainForm);
					}
					handled = true;
					break;

				/* Changes the memo category */
				case popDetailsCategory:

					/* Clear the field color settings */

					SetFieldColor(blackColor);

					categoryUpdated = CategorySelect(dbMemoPadRef, FrmGetActiveForm(),
									popDetailsCategory, lstDetailsCategory,
									false, &category, categoryName, 1, 0);

					/* Restore the field color setting as it was before */

					SetFieldColor(currentFieldColor);

					categoryEdited = (categoryUpdated || categoryEdited);
					handled = true;
			}
			break;

		case frmOpenEvent:
			DetailsFormInit();
			FrmDrawForm(FrmGetActiveForm());
			category = currentCategory;
			categoryEdited = false;
			handled = true;
			break;
	}

	return handled;
}

