#include "FontForm.h"
#include "Database.h"
#include "Utils.h"

static void FontFormApplyChanges(void) FONT_SECTION;
static void FontFormSelectColor(Boolean) FONT_SECTION;
static void FontFormDrawRectangleWithColor(IndexedColorType, UInt16) FONT_SECTION;
static Boolean PickColor(IndexedColorType *, DmResID) FONT_SECTION;
static void FontAndColorInit(FontID *, IndexedColorType *, UInt8, UInt16) FONT_SECTION;
static void FontFormInit(void) FONT_SECTION;

static Boolean fontForm;
static Boolean colorChanged = false;
static IndexedColorType newTableColor;

static UInt16 updateCode = 0;

static void FontFormInit(void)
{
	if (fontForm == fontRecord)
		FontAndColorInit(&currentFieldFont, &currentFieldColor, grpFontFieldFontGroup, gdgFontFieldColor);


	if ((fontForm == fontList) && (useSingleFont == separateFont)) return;

	FontAndColorInit(&currentTableFont, &currentTableColor, grpFontTableFontGroup, gdgFontTableColor);
}

/************************************************
 * FontAndColorInit:
 * Set the correct font button and
 * the correct color.
 ************************************************/

static void FontAndColorInit(FontID * currentFontP, IndexedColorType * currentColorP, 
	UInt8 grpSelection, UInt16 objectID)
{
	UInt16 controlID;
	FormPtr formP;
	RectangleType tempRect;
	Boolean grpField;

	formP = FrmGetActiveForm();

	grpField = (grpSelection == grpFontFieldFontGroup);

	/* Updates the currently selected field font */
	switch (*currentFontP) {
		case stdFont:
			if (grpField)
				controlID = pshFontSmallFieldFont;
			else controlID = pshFontSmallTableFont;
			break;

		case boldFont:
			if (grpField)
				controlID = pshFontBoldFieldFont;
			else controlID = pshFontBoldTableFont;
			break;

		case largeFont:
			if (grpField)
				controlID = pshFontLargeFieldFont;
			else controlID = pshFontLargeTableFont;
			break;

		case largeBoldFont:
			if (grpField)
				controlID = pshFontLargeBoldFieldFont;
			else controlID = pshFontLargeBoldTableFont;
			break;
	}

	FrmSetControlGroupSelection(formP, grpSelection, controlID);

	/* Update the currently selected field color or
	 * display a message if the PalmOS is under PalmOS35
	 * and color is not supported. */

	if (OSVersion < OS35) {
		FntSetFont(stdFont);

		if (objectID == gdgFontTableColor) {
			FrmShowObject (formP, FrmGetObjectIndex(formP, lblFontTableColorNotSupported));
		} else FrmShowObject (formP, FrmGetObjectIndex(formP, lblFontFieldColorNotSupported));
	} else {
		FontFormDrawRectangleWithColor(*currentColorP, objectID);
	}
}

/************************************************
 * FontFormDrawRectangleWithColor:
 * Draw a rectangle frame filled in
 * with the color passed as parameter.
 ************************************************/

static void FontFormDrawRectangleWithColor(IndexedColorType color, UInt16 objectID)
{
	RectangleType tempRect;
	FormPtr frmP;

	frmP = FrmGetActiveForm();

	// Save the current draw state values
	WinPushDrawState();

	FrmGetObjectBounds(frmP, FrmGetObjectIndex(frmP, objectID), &tempRect);

	// Fill the interior with the correct color
	WinSetForeColor(color);
	WinDrawRectangle(&tempRect, 0);

	// Draw the frame
	WinSetForeColor(blackColor);
	WinDrawRectangleFrame(rectangleFrame, &tempRect);

	// Restore the draw state values to the last saved
	WinPopDrawState();
}

/************************************************
 * PickColor:
 * Display the UI color palette
 * and the correct title.
 * return the color selected.
 ************************************************/

static Boolean PickColor(IndexedColorType * newColor, DmResID resID)
{
	RGBColorType rgb;
	MemHandle strPickColorH;
	Char * strPickColorP;
	Boolean colorSelected;

	// Set the correct title in the pick color form.
	strPickColorH = DmGetResource ('tSTR', resID);
	strPickColorP = MemHandleLock (strPickColorH);

	colorSelected = UIPickColor (newColor, &rgb, UIPickColorStartPalette, strPickColorP, NULL);
	
	MemHandleUnlock(strPickColorH);
	DmReleaseResource(strPickColorH);

	return (colorSelected);
}

/************************************************
 * FontFormSelectColor:
 * Select a color with the
 * system palette.
 ************************************************/

static void FontFormSelectColor(Boolean item)
{
	IndexedColorType newFieldColor;
	Boolean colorSelected;

	newTableColor = currentTableColor;
	newFieldColor = currentFieldColor;

	if ((fontForm == fontList) && (useSingleFont == separateFont))
		 newTableColor = blackColor;

	if (item == colorTable)
		colorSelected = PickColor(&newTableColor, strPickTableColor);
	else colorSelected = PickColor(&newFieldColor, strPickFieldColor);

	if ((newTableColor != currentTableColor) && colorSelected) {
			currentTableColor = newTableColor;
			colorChanged = true;

			/* Update the currently selected table color */
			FontFormDrawRectangleWithColor(currentTableColor, gdgFontTableColor);

			updateCode = updateFontChanged;
	}

	if (fontForm == fontRecord) {
		/* If the color field has changed,
		 * update the edit form when return.
		 */
		if (newFieldColor != currentFieldColor) {
			currentFieldColor = newFieldColor;
			colorChanged = true;

			/* Update the currently selected field color */
			FontFormDrawRectangleWithColor(currentFieldColor, gdgFontFieldColor);

			updateCode = updateFontChanged;
		}
	}
}

/************************************************
 * FontFormApplyChanges:
 * Change the font and color settings
 * of the field and the table according
 * to the user choice.
 ************************************************/

static void FontFormApplyChanges(void)
{
	UInt16 controlID;
	UInt16 itemNumber;
	FormPtr formP;
	FontID newTableFont, newFieldFont;
	Boolean fontChanged = false;

	formP = FrmGetActiveForm();

	/* Get the currently selected table font */
	itemNumber = FrmGetControlGroupSelection(formP, grpFontTableFontGroup);

	if (itemNumber != frmNoSelectedControl) {
		controlID = FrmGetObjectId (formP, itemNumber);

		switch (controlID) {
			case pshFontSmallTableFont:
				newTableFont = stdFont;
				break;

			case pshFontBoldTableFont:
				newTableFont = boldFont;
				break;

			case pshFontLargeTableFont:
				newTableFont = largeFont;
				break;

			case pshFontLargeBoldTableFont:
				newTableFont = largeBoldFont;
				break;
		}

		if (newTableFont != currentTableFont) {
				currentTableFont = newTableFont;
				updateCode = updateFontChanged;
		}
	}

	if (fontForm == fontRecord) {
		/* Get the currently selected field font */
		controlID = FrmGetObjectId (formP,
			FrmGetControlGroupSelection(formP, grpFontFieldFontGroup));
	
		switch (controlID) {
			case pshFontSmallFieldFont:
				newFieldFont = stdFont;
				break;

			case pshFontBoldFieldFont:
				newFieldFont = boldFont;
				break;

			case pshFontLargeFieldFont:
				newFieldFont = largeFont;
				break;

			case pshFontLargeBoldFieldFont:
				newFieldFont = largeBoldFont;
				break;
		}

		/* If the font field has changed,
		 * update the edit form when return.
		 */
		if (newFieldFont != currentFieldFont) {
			currentFieldFont = newFieldFont;
			updateCode = updateFontChanged;
		}
	} else {
		if (useSingleFont == separateFont) {
			/* If we are using the separate font feature
			 * and if user selected a text color or a font,
			 * update all the records with the new settings.
			 */
			UInt16 recordNum = 0;
			MemHandle recordH;

			if (itemNumber != frmNoSelectedControl) fontChanged = true;

			if (fontChanged || colorChanged) {

				/* Display a alert to confirm */
				if (FrmAlert(altChangeTableFontOrColorOfCategoryAlert) == 0) {
					while (true) {
						recordH = DmQueryNextInCategory (dbCryptoPadRef, &recordNum, currentCategory);
 						if (!recordH) break;

 						GetRecordFont(recordNum);

	 					if (fontChanged) currentTableFont = newTableFont;
 						if (colorChanged) currentTableColor = newTableColor;

		 				SetRecordFont(recordNum);

	 					recordNum++;
					}
				}
			}
		}
	}
}


Boolean FontFormHandleEvent(EventPtr eventP)
{
	Boolean handled = false;
	FormPtr frmP;
	RectangleType fieldColorRect, tableColorRect;

	static returnForm;

	switch (eventP->eType) {
		case fontProcessEvent:
			fontForm = eventP->data.generic.datum[1];
			/* Set the correct form to return to
			 * according to the event data.
			 */
			if (fontForm == fontRecord)
				returnForm = frmEditForm;
			else returnForm = frmMainForm;
			FrmDrawForm(FrmGetActiveForm());
			FontFormInit();
			break;

		case penDownEvent:
			if (OSVersion < OS35) break;
			frmP = FrmGetActiveForm();
			
			// tapped on the card area?
			FrmGetObjectBounds(frmP, FrmGetObjectIndex(frmP, gdgFontTableColor), &tableColorRect);
			
			if (RctPtInRectangle(eventP->screenX, eventP->screenY, &tableColorRect)) {
				FontFormSelectColor(colorTable);
				handled = true;
			} else if (fontForm == fontRecord) {
				FrmGetObjectBounds(frmP, FrmGetObjectIndex(frmP, gdgFontFieldColor), &fieldColorRect);

				if (RctPtInRectangle(eventP->screenX, eventP->screenY, &fieldColorRect)) {
					FontFormSelectColor(colorField);
					handled = true;
				}
			}
			break;

		case ctlSelectEvent:
			switch (eventP->data.ctlSelect.controlID) {
				case btnFontOK:
					FontFormApplyChanges();
					FrmReturnToForm(returnForm);
	
					/* if the current category has be deleted, renamed or
					 * merged with another category, we need to update the
					 * edit form.
					 */
					if (updateCode)
						FrmUpdateForm (returnForm, updateCode);			
					handled = true;
					break;

				case btnFontCancel:
					FrmReturnToForm(returnForm);
					handled = true;
					break;
		}
			break;
	}

	return handled;
}

