/* CryptoPad : a encryption/decryption enabled memopad replacement
 *
 * Maxime Labelle & Jean-Baptiste Labelle - 2000
 */

#include "MainForm.h"
#include "EditForm.h"
#include "DetailsForm.h"
#include "PasswordForm.h"
#include "FontForm.h"
#include "QueryGlobalPasswordForm.h"
#include "SetAndChangeGlobalPasswordForm.h"
#include "PrefsForm.h"
#include "Database.h"
#include "blowfish.h"

static Boolean ApplicationHandleKeyDownEvent(EventPtr);
//static Boolean ApplicationHandleImportantMemo(EventPtr);
static WChar GetCryptoPadHardKey(void);

/******************************************
 * Blowfish:
 * Main encryption/decryption entry point
 ******************************************
 */

void Blowfish(Char * input, Char * output, Char * password, UInt16 nLength, Int16 encrypt)
{
	BF_KEY* key;
	Int16 num;
	UInt8 iv[8]= {0xfe,0xdc,0xba,0x98,0x76,0x54,0x32,0x10};

	key = (BF_KEY*) MemPtrNew(sizeof(BF_KEY));
	ErrFatalDisplayIf(key == NULL, "Unable to expand key.");

	BF_set_key(key, StrLen(password), password);

	num = 0;
	BF_cfb64_encrypt(input, output, nLength, key, iv, &num, encrypt);
	output[nLength] = '\0';

	MemPtrFree(key);
}

/***********************************
 * RomVersionCompatible
 * Check that the ROM version meets the
 * minimum requirement.                    
 ***********************************/
static Boolean RomVersionCompatible (UInt32 requiredVersion)
{
	UInt32 romVersion;

	/* See if we're on in minimum required version of the ROM or later */
	FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);

	if (romVersion < requiredVersion) {
		FrmAlert (altRomIncompatibleAlert);
		return (false);
	}

	return (true);
}

Boolean ApplicationHandleEvent(EventPtr eventP)
{
	Boolean handled = false;
	Int16 formID;
	FormPtr formP;

	if (eventP->eType == frmLoadEvent) {

		// Load the form and assign its event handler
		formID = eventP->data.frmLoad.formID;
		formP = FrmInitForm(formID);
		FrmSetActiveForm(formP);

		switch (formID) {
			case frmMainForm:
				FrmSetEventHandler(formP, MainFormHandleEvent);
				break;

			case frmEditForm:
				FrmSetEventHandler(formP, EditFormHandleEvent);
				break;

			case frmDetailsForm:
				FrmSetEventHandler(formP, DetailsFormHandleEvent);
				break;

			case frmFontForm:
			case frmTableFontForm:
				FrmSetEventHandler(formP, FontFormHandleEvent);
				break;

			case frmPasswordForm:
				FrmSetEventHandler(formP, PasswordFormHandleEvent);
				break;

			case frmQueryGlobalPasswordForm:
				FrmSetEventHandler(formP, QueryGlobalPasswordFormHandleEvent);
				break;

			case frmSetGlobalPasswordForm:
				FrmSetEventHandler(formP, SetAndChangeGlobalPasswordFormHandleEvent);
				break;

			case frmConfirmGlobalPasswordForm:
				FrmSetEventHandler(formP, SetAndChangeGlobalPasswordFormHandleEvent);
				break;

			case frmChangeGlobalPasswordForm:
				FrmSetEventHandler(formP, SetAndChangeGlobalPasswordFormHandleEvent);
				break;

			case frmPrefsForm:
				FrmSetEventHandler(formP, PrefsFormHandleEvent);
				break;
		}

		handled = true;
	}

	return handled;
}

/************************************************
 * GetApplicationPrefs:
 * Load the preferences and do any fixups needed for 
 * backwards and forwards compatibility
 ************************************************/
void GetApplicationPrefs(UInt32* currentRecordID)

{
	ApplicationPreferenceType appPrefs;
	UInt16 prefSize;
	Int16 result;
	RGBColorType rgb;

	/* Define the blackColor variable */

	if (OSVersion >= OS35) {
		MemSet(&rgb, sizeof(rgb),0);
		blackColor = WinRGBToIndex(&rgb);
	}	

	prefSize = sizeof(ApplicationPreferenceType);

	result = PrefGetAppPreferences(cryptoPadCreator, dbPrefID, &appPrefs,
												&prefSize, true);

	if (dbVersion <= result && result != noPreferenceFound) {
		currentForm      = appPrefs.currentForm;
		currentCategory  = appPrefs.currentCategory;
		currentFieldFont = appPrefs.currentFieldFont;

		currentRotateOrder	= appPrefs.currentRotateOrder;

		currentRecord    = appPrefs.currentRecord;
		topRecord        = appPrefs.topRecord;
		currentTableFont = appPrefs.currentTableFont;

		moveRecordOnTop	= appPrefs.moveRecordOnTop;

		editScrollPosition = appPrefs.editScrollPosition;
		showLines		 = appPrefs.showLines;
		saveBackup		 = appPrefs.saveBackup;
		showAllCategories  = appPrefs.showAllCategories;
		useSingleFont	= appPrefs.useSingleFont;
		*currentRecordID  = appPrefs.currentRecordID;
		importantRecordID  = appPrefs.importantRecordID;
		currentTableColor = appPrefs.currentTableColor;
		currentFieldColor = appPrefs.currentFieldColor;
		useLookupField = appPrefs.useLookupField;
	} 

	else {
		/* Update the records from CryptoPad to make them
		 * conforme with the new CryptoRecordType.
		 */
		UpdateCryptoPadDB();

		/* Version 3.7 does not use anymore the 
		 * currentDB and the topDB because the
		 * record type have changed.
		 * The application preferences couldn't be
		 * retrieved. Try version 3.5 preferences instead. */
		
		*currentRecordID  	= noRecordSelectedID;
		importantRecordID  	= noRecordSelectedID;
	 	currentRecord   	= noRecordSelected;
		topRecord        	= 0;
		currentTableColor	= blackColor;
		currentFieldColor	= blackColor;
		useLookupField		= false;
				
		prefSize = sizeof(ApplicationPreferenceTypeV36);
		result = PrefGetAppPreferences(cryptoPadCreator, dbPrefIDV36, &appPrefs,
			&prefSize, true);
										
		if (dbNewOptionsVersion <= result && result != noPreferenceFound) {
			currentForm      = appPrefs.currentForm;
			currentCategory  = appPrefs.currentCategory;
			currentFieldFont = appPrefs.currentFieldFont;

			currentRotateOrder	= appPrefs.currentRotateOrder;

			currentTableFont = appPrefs.currentTableFont;

			moveRecordOnTop	= appPrefs.moveRecordOnTop;

			editScrollPosition = appPrefs.editScrollPosition;
			showLines		 = appPrefs.showLines;
			saveBackup		 = appPrefs.saveBackup;
			showAllCategories  = appPrefs.showAllCategories;
			useSingleFont	= appPrefs.useSingleFont;
		}
				
		else {
			/* Start with default values */
			currentForm     	= frmMainForm;
			currentRecord   	= noRecordSelected;
			currentCategory 	= dmAllCategories;
			currentFieldFont 	= stdFont;
			currentTableFont 	= stdFont;
			topRecord  		    = 0;
			showLines			= linesShown;
			editScrollPosition 	= 0;
    		saveBackup			= true;
			showAllCategories 	= true;
			useSingleFont 		= singleFont;
			*currentRecordID  = noRecordSelectedID;
			importantRecordID  = noRecordSelectedID;
			currentRotateOrder	= roAll;
			moveRecordOnTop		= recNotMove;
			currentTableColor	= blackColor;
			currentFieldColor	= blackColor;
			useLookupField		= false;
		}
	}
}

/************************************************
 * SetApplicationPrefs:
 * Save the preferences and do any fixups needed for 
 * backwards and forwards compatibility
 ************************************************/
void SetApplicationPrefs(UInt16 scrollPosition)
{
	ApplicationPreferenceType appPrefs;
	UInt32 uniqueID;

	/* Save the current settings for later
	 * execution of the application.
	 */
	appPrefs.currentForm      	= currentForm;
	appPrefs.currentCategory  	= currentCategory;
	appPrefs.currentFieldFont 	= currentFieldFont;
	appPrefs.currentRecord    	= currentRecord;
	appPrefs.topRecord        	= topRecord;
	appPrefs.currentTableFont 	= currentTableFont;
	appPrefs.editScrollPosition = scrollPosition;
	appPrefs.showLines 		  	= showLines;
	appPrefs.saveBackup		  	= saveBackup;
	appPrefs.showAllCategories 	= showAllCategories;
	appPrefs.useSingleFont	  	= useSingleFont;
	appPrefs.currentRotateOrder = currentRotateOrder;
	appPrefs.moveRecordOnTop	= moveRecordOnTop;
	appPrefs.currentTableColor 	= currentTableColor;
	appPrefs.currentFieldColor 	= currentFieldColor;
	appPrefs.importantRecordID 	= importantRecordID;
	appPrefs.useLookupField	= useLookupField;

	// Get the current record's unique id and save it with the state 
	// information.
	if ((!isEncrypted(currentRecord, dbCryptoPadRef)) && DmQueryRecord (dbCryptoPadRef, currentRecord) != 0) {
			DmRecordInfo (dbCryptoPadRef, currentRecord, NULL, &uniqueID, NULL);
			appPrefs.currentRecordID = uniqueID;
	 } else {
		appPrefs.currentRecordID = noRecordSelectedID;
		appPrefs.currentRecord = noRecordSelected;
	}

	PrefSetAppPreferences(cryptoPadCreator, dbPrefID, dbVersion,
							&appPrefs, sizeof(ApplicationPreferenceType), true);
}

/************************************************
 * ChangeGeneralColor:
 * Change the screen depth.
 * Display some UI controls in grey
 * and change the color of several item
 * in the UI interface.
 ************************************************/

static ChangeGeneralColor(void)
{
	RGBColorType darkGrey;
	RGBColorType clearGrey;
	RGBColorType middleGrey;
	UInt32 depth;

	darkGrey.index = 0;
	darkGrey.r = darkGrey.g = darkGrey.b = 50;

	clearGrey.index = 0;
	clearGrey.r = clearGrey.g = clearGrey.b = 230;

	middleGrey.index = 0;
	middleGrey.r = middleGrey.g = middleGrey.b = 160;

	/* Retrieve the supported screen depth. */
	WinScreenMode(winScreenModeGetSupportedDepths, NULL, NULL, &depth, NULL);

	if (depth == 0x808B)
		depth = 16;
		else if (depth == 0x8B)
			depth = 8;
			else if (depth == 0x0B)
				depth = 4;
				else if (depth == 0x03)
					depth = 2;
				else depth = 1;

	if (depth == 4) {
		/* Set the screen depth. */
		WinScreenMode(winScreenModeSet, NULL, NULL, &depth, NULL);

		UIColorSetTableEntry (UIMenuSelectedFill, &middleGrey);
		UIColorSetTableEntry (UIMenuSelectedForeground, &clearGrey);

		UIColorSetTableEntry (UIFieldTextLines, &middleGrey);
		UIColorSetTableEntry (UIFieldCaret, &darkGrey);
	}
}



/* Standard routines */

static Err StartApplication(void)
{
	UInt16 recAttributes;
	UInt16 mode;
	UInt32 uniqueID;
	UInt32 currentRecordID = 0;
	Boolean recordFound;

	if (!RomVersionCompatible (version))
	return 1;

	FtrGet(sysFtrCreator, sysFtrNumROMVersion, &OSVersion);

	// Determime if secret record should be shown.

	if (OSVersion < OS35) {
		if (PrefGetPreference(prefHidePrivateRecordsV33))
			PrivateRecordVisualStatus = hidePrivateRecords;
		else PrivateRecordVisualStatus = showPrivateRecords;
	} else {
		PrivateRecordVisualStatus = (privateRecordViewEnum)PrefGetPreference(prefShowPrivateRecords);

		/* Change the screen depth */
		ChangeGeneralColor();
	}

	if (PrivateRecordVisualStatus == hidePrivateRecords)
		mode = dmModeReadWrite;
	else mode = dmModeReadWrite | dmModeShowSecret;

	dbCryptoPadRef = OpenOrCreateCryptoDatabase(mode);
	dbMemoPadRef = OpenMemoDatabase(mode);

	SynchroniseCategories();

	GetPrefsGlobalPassword();

	/* Read the application preferences */

	GetApplicationPrefs(&currentRecordID);

	/* Check if records have been added/removed
	 * in CryptoPad or while a hotsync.
	 */

	SynchroniseRecord();

	 if (GetMemoPadSortOrder(dbMemoPadRef) == soAlphabetic) 
	 	DmQuickSort(dbCryptoPadRef,
			(DmComparF*) &CompareCryptoPadRecords, soAlphabetic);

	/* Check that the category still exists; it
	 * may have been deleted after a hotsync.
	 */

	CategoryGetName(dbMemoPadRef, currentCategory, categoryName);
	if (*categoryName == 0) {
		currentCategory = dmAllCategories;
		showAllCategories = true;
	}

	/* Count the records in both databases */

	if (showAllCategories)
		memosInCategory = DmNumRecordsInCategory (dbCryptoPadRef, dmAllCategories);
	else memosInCategory = DmNumRecordsInCategory (dbCryptoPadRef, currentCategory);

	/* Start watching the button pressed to get into this app.  If it's held down
	 * long enough then we need to go into the important memo.
	 */
//	tickAppButtonPushed = TimGetTicks();
	
	/* Mask off the key to avoid repeat keys causing clicking sounds */
//	KeySetMask(~KeyCurrentState());

	/* If we have to display the mainForm, don't worry about
	 * the previously selected record.
	 */

//	if (currentForm == frmMainForm) return 0;


	/* Check that the current record still exists and hasn't
	 * been deleted as a result of a hotsync.
	 */

	if (DmQueryRecord (dbCryptoPadRef, currentRecord) != 0) {
		DmRecordInfo (dbCryptoPadRef, currentRecord, &recAttributes, &uniqueID, NULL);
		recordFound = (uniqueID == currentRecordID) && 
			((PrivateRecordVisualStatus == showPrivateRecords) || 
			(!(recAttributes & dmRecAttrSecret)));
	}

	if (! recordFound) {
		topRecord = 0;
		currentRecord = noRecordSelected;
		currentForm = frmMainForm;
		editScrollPosition = 0;
		}

	return 0;
}

static void EventLoop(void)
{
	EventType event, customEvent;
	UInt16 error;
	Boolean memoPadButtonSeen = false;

	do {
//		EvtGetEvent(&event, (tickAppButtonPushed == 0) ? evtWaitForever :2);
		EvtGetEvent(&event, evtWaitForever);

//		if (!ApplicationHandleImportantMemo(&event))
			memoPadButtonSeen = ApplicationHandleKeyDownEvent(&event);

		if (!SysHandleEvent(&event))
			if (!MenuHandleEvent(0, &event, &error))
				if (!ApplicationHandleEvent(&event))
					FrmDispatchEvent(&event);
		
		if (memoPadButtonSeen == true) {
		/* If we get here, then we had the MemoPad button pressed and SysHandleEvent decided
		 * to handle it itself. BUT if AppHack had intercepted the button, AppHack did nothing
		 * and we will miss the button completely, so....
		 * Queue up a custom event to switch to the next view. If SysHandleEvent is all set to
		 * go to another application, we may not get here at all, but if we do, this event will
		 * be ignored by any other application, so we should be ok (unless that app also has
		 * custom events, but that's why we biased the event number by 143 to avoid that).
		 */

			// Send an event to load the form we want to goto.
    		customEvent.eType = AppHackBugEvent;
  			customEvent.data.generic.datum[1] = true;
		    EvtAddEventToQueue (&customEvent);
		    memoPadButtonSeen = false;
		}

	} while (event.eType != appStopEvent);
}


/**************************************************************
 * GetCryptoPadHardKey:
 * Return the hard key number that launch CryptoPad.
 **************************************************************/

static WChar GetCryptoPadHardKey(void) {
	if (PrefGetPreference(prefHard1CharAppCreator) == cryptoPadCreator) return hard1Chr;
		else if (PrefGetPreference(prefHard2CharAppCreator) == cryptoPadCreator) return hard2Chr;
			else if (PrefGetPreference(prefHard3CharAppCreator) == cryptoPadCreator) return hard3Chr;
				else if (PrefGetPreference(prefHard4CharAppCreator) == cryptoPadCreator) return hard4Chr;
					else return (-1);
}

/**************************************************************
 * ApplicationHandleKeyDownEvent:
 * Handle the hard keys pressed.
 **************************************************************/

static Boolean ApplicationHandleKeyDownEvent(EventPtr eventP)
{
	Boolean memoPadButtonSeen = false; // Until we know better...
	EventType customEvent;

	if (eventP->eType == keyDownEvent) {
		/* We have a kludge here to fix a bug in AppHack - it steals the MemoPad
		 * button completely if CryptoPad is running!
		 */

		if ((eventP->data.keyDown.modifiers & commandKeyMask) 
			&& (eventP->data.keyDown.chr == GetCryptoPadHardKey()))
//			&& (goneIntoImportantMemo == false))
					memoPadButtonSeen = true;

 		/* If the Palm is opened with the MemoPad hard key,
 		 * we send this custom event with the parameter to false
 		 * to prevent the current form handle event function to
 		 * change of category.
 		 */

		if (eventP->data.keyDown.modifiers & poweredOnKeyMask) {
			customEvent.eType = AppHackBugEvent;
	    	customEvent.data.generic.datum[1] = false;
		    EvtAddEventToQueue (&customEvent);
		}

		/* The Palm shut off so if we were in an encrypted memo,
		 * we must exit and reset the global password entered
		 * so you user has to reenter it to accesse the memo.
		 */

		if ((eventP->data.keyDown.chr == hardPowerChr) || 
			(eventP->data.keyDown.chr == autoOffChr)) {
			if (wasEncrypted && (currentPasswordP != 0)) {
				FrmCloseAllForms();
				FrmGotoForm(frmMainForm);
			}
		globalPasswordP = 0;
		}
	}


	return memoPadButtonSeen;
}


/**************************************************************
 * ApplicationHandleImportantMemo:
 * Handle the hard key pressed for a few seconds
 * for the Important memo feature.
 **************************************************************/
/*
static Boolean ApplicationHandleImportantMemo(EventPtr eventP)
{
	Boolean memoPadButtonSeen = false; // Until we know better...
	EventType customEvent;
	UInt16 recordNum;
	Err err;
*/
	// Check if a button being held down is released
//	if (tickAppButtonPushed != 0) {
   		// This is the case when the button is let up
/*   		if ((KeyCurrentState() & (keyBitHard1 | keyBitHard2 | keyBitHard3 | keyBitHard4)) == 0) {
		   	if (goneIntoImportantMemo) {
		   		goneIntoImportantMemo = false;
	      	
		      	tickAppButtonPushed = 0;
*/	      	
				// Allow the masked off key to now send keyDownEvents.
//				KeySetMask(keyBitsAll);

//	   		} else if (eventP->eType == nilEvent) {
		      	// Send the keyDownEvent to the app.  It was stripped out
		      	// before but now it can be sent over the nullEvent.  It
	    	  	// may be nullChr from when the app was launched.  In that case
	      		// we don't need to send the app's key because the work expected,
		      	// which was switching to this app, has already been done.
/*		      if (appButtonPushed != nullChr) {
			      	eventP->eType = keyDownEvent;
			      	eventP->data.keyDown.chr = appButtonPushed;
			      	eventP->data.keyDown.modifiers = appButtonPushedModifiers;
		      	}
	      	
	      		tickAppButtonPushed = 0;
*/	      	
				// Allow the masked off key to now send keyDownEvents.
/*				KeySetMask(keyBitsAll);
	      	}
		}
*/		// This is the case when the button is depresed long enough to go into the important memo
/*		else if (tickAppButtonPushed + sysTicksPerSecond <= TimGetTicks() &&
				!goneIntoImportantMemo) {
			goneIntoImportantMemo = true;

			err = DmFindRecordByID (dbCryptoPadRef, importantRecordID, &recordNum);

			if ((err == dmErrUniqueIDNotFound) || (DmQueryRecord(dbCryptoPadRef, recordNum) == 0)) {
				FrmAlert(altNoImportantMemoIDAlert);
				FrmGotoForm(frmMainForm);
			} else {
				UInt16 recAttributes;

				if (err) return false;

			FrmGotoForm(frmEditForm);

				DmRecordInfo (dbCryptoPadRef, recordNum, &recAttributes, NULL, NULL);
*/

			// If this is a "private" record, then determine what is to be shown.
/*
				if (recAttributes & dmRecAttrSecret) {
					switch (PrivateRecordVisualStatus) {
						case showPrivateRecords:
							MainFormEnterTableItem();
							break;
					
						case maskPrivateRecords:
							if (SecVerifyPW (showPrivateRecords) == true) {
								PrivateRecordVisualStatus = showPrivateRecords;
								MainFormEnterTableItem();
							}
							break;

						case hidePrivateRecords:
							FrmAlert(altImportantMemoIsPrivate);
							return;
					}
				} else MainFormEnterTableItem();

				eventP->eType = nilEvent;
				EvtAddEventToQueue (eventP);
*/
				/* Updates the currently selected category */
/*
				currentCategory = (recAttributes & dmRecAttrCategoryMask);
				ChangeCategory(currentCategory);

			}
		}
	}
	   
    else if (eventP->eType == keyDownEvent) {
		if ((eventP->data.keyDown.chr == GetCryptoPadHardKey())
			&& TxtCharIsHardKey(eventP->data.keyDown.modifiers, eventP->data.keyDown.chr)
			&& !(eventP->data.keyDown.modifiers & autoRepeatKeyMask)) {
*/			// Remember which hard key is mapped to CryptoPad
			// because it may need to be sent later.
/*			appButtonPushed = eventP->data.keyDown.chr;
			appButtonPushedModifiers = eventP->data.keyDown.modifiers;
		
			tickAppButtonPushed = TimGetTicks();
*/			
			// Mask off the key to avoid repeat keys causing clicking sounds
//			KeySetMask(~KeyCurrentState());

			// Don't process the key
/*			return true;
		}
	}	

	return false;
}
*/
static void StopApplication(void)

{
	UInt16 scrollPosition = 0;
	FormPtr formP;
	FieldPtr fieldP;

	// If we are in the Edit form, get the current scroll position.	

	if ((currentForm == frmEditForm) && (currentRecord != noRecordSelected)) {
		formP = FrmGetFormPtr (frmEditForm);
		fieldP = FrmGetObjectPtr (formP, FrmGetObjectIndex (formP, fldEditText));
		scrollPosition = FldGetScrollPosition (fieldP);
	}

	// Close all open forms,  this will force any unsaved data to 
	// be written to the database.

	FrmCloseAllForms();

	/* Get the current record's unique id and save it with the state 
	 * information.
	 */

	SetPrefsGlobalPassword(dbCryptoPadRef);	

	/* Save the current settings for lat
	 * execution of the application.
	 */

	SetApplicationPrefs(scrollPosition);

	if (OSVersion >= OS35) {
		UInt32 depth;

		/* Get back to the current depth screen state
		 * that is : balck and white as others applications
		 * must do that themselves.
		 */
	
		depth = 1;
		WinScreenMode(winScreenModeSet, NULL, NULL, &depth, NULL);
	}

	DmCloseDatabase(dbMemoPadRef);
	DmCloseDatabase(dbCryptoPadRef);
}

UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
	Err err;
	Boolean appIsActive;

	if (cmd == sysAppLaunchCmdNormalLaunch) {
		err = StartApplication();
		if (err) return (1);

		FrmGotoForm(currentForm);

		EventLoop();
		StopApplication();
	}

	else if (cmd == sysAppLaunchCmdFind) {
		appIsActive = launchFlags & sysAppLaunchFlagSubCall;
		Search ((FindParamsPtr)cmdPBP, appIsActive);
		}
	
	
	/* This action code might be sent to the app when it's already running
	*  if the use hits the Find soft key next to the Graffiti area.
	*/
	else if (cmd == sysAppLaunchCmdGoTo) {
		if (launchFlags & sysAppLaunchFlagNewGlobals) 
			{
			err = StartApplication ();
			if (err) return (1);
			GoToItem ((GoToParamsPtr) cmdPBP, true);
			
			EventLoop ();
			StopApplication ();
		} else {
			/* The app was already running.  Open a form to display the record */
			GoToItem ((GoToParamsPtr) cmdPBP, false);
			}
		}
	/* Launch code sent to running app to save all data.
	* This is neccessary before a find for some applications because
	* some apps do not edit in place and the user might search for
	* data not saved.  Or saving a record might reorder the database.
	* Saving before a find avoids finding a record, saving, resorting,
	* and then displaying a record moved to the position of the found
	* record.
	*/
//	else if (cmd == sysAppLaunchCmdSaveData) {
//		FrmSaveAllForms ();
//		}

	return 0;
}
