// $Header: /cvsroot/cryptopad/cryptopad_classic/v4/src/MainForm.c,v 1.5 2005/10/10 05:12:38 jbauman Exp $

// This needs to be included before PalmUtils.h, or #defines there hose
// declarations in the stuff included here.
#include <palmOne_68K.h>
#include <68K/System/HsExt.h>

#include "MainForm.h"
#include "Database.h"
#include "Utils.h"
#include "blowfish.h"
#include "Beam.h"
#include <PalmUtils.h>

#include <SysEvtMgr.h>


static Boolean MainFormSeekRecord (UInt16* , Int16, Int16);
static WinHandle MainFormInvertMoveIndicator(RectanglePtr, WinHandle, Boolean);
static void MainFormDragMemo(EventPtr);
static void MainFormEditRecord(TablePtr tableP, Int16 row);
static void MainFormSelectTableItem (Boolean, TablePtr, Int16, Int16, RectangleType *r);
static void ReplaceTwoColors (const RectangleType *, UInt16,
			UIColorTableEntries, UIColorTableEntries, UIColorTableEntries, UIColorTableEntries);

static void MainFormInit(void);

static void MainFormDrawTableItem(void *, Int16, Int16, RectanglePtr);
static void DrawCryptoPadRecordText(UInt16, UInt16, RectanglePtr);
static void DrawMemoPadRecordText(UInt16, UInt16, RectanglePtr);
static void MainFormDisplayMask (RectanglePtr);
static Int16 MainFormNumberOfRows (TablePtr);

static void MainFormLoadTable(void);
static void MainFormLoadRecords(void);
static void MainFormShowFullDisplay(void);

static void EncryptCategory();
static void DecryptCategory();
static ProgressInfoPtr ProgressFormInitAndDraw (FormPtr frmP);
static Boolean ProgressCallback (void* progressInfoP, ProgrType* progressP, Char* messageP);
static void ProgressUpdateGauge (RectangleType*	gaugeRP, UInt32 gaugeLevel);
static UInt32 ProgressHandleEvent (Int32 timeout);

static void MainFormSelectCategory(void);
static void MainFormNextCategory(void);
static UInt16 MainFormGetNextCategory(UInt16);

static void MainFormUpdateScrollBar(void);
static Boolean MainFormUpdateDisplay(UInt16);
static Boolean MainFormScrollLines(Int16);
static Boolean MainFormScrollPage(WinDirectionType);
static Boolean MainFormHandleListChoices(UInt8, UInt16);

static void MainFormHandleMenuEvent(UInt16);

static Boolean RecordLookupString(DmOpenRef, Char *, UInt16, UInt16*, Boolean *);
static Boolean MainFormLookupString (EventPtr);
static void MainFormSelectRecord (UInt16, TablePtr);
static void MainFormClearLookupString (void);
static void MainFormDisplayLookupField(void);

static Boolean HighlightTable(FormPtr formP, Boolean addHighlight);
static Boolean HighlightObject(FormPtr formP, UInt16 object);
static Boolean RowNotLast(TablePtr tableP, Int16 row);
// static void SetLookupSkip(FormPtr formP, Boolean no_skip);

// Newly added for 5-way navigation support as of v 4.11
static Boolean tblInteractive = false;
static Int16 currentRow = 0;
static Boolean didCategoryChange = false;

/***************************************
 * MainFormSeekRecord:
 * utility routine for MainFormDragMemo
 ***************************************
 */

static Boolean MainFormSeekRecord (UInt16* indexP, Int16 offset, Int16 direction)
{
	DmSeekRecordInCategory (dbCryptoPadRef, indexP, offset, 
		direction, currentCategory);

	if (DmGetLastErr()) return (false);
	
	return (true);
}


/***************************************
 * MainFormInvertMoveIndicator:
 * Saves the screen behind the move
 * indicator while dragging a record.
 ***************************************
 */

static WinHandle MainFormInvertMoveIndicator(RectanglePtr itemR,
			WinHandle savedBits, Boolean draw)
{
	WinHandle winH = 0;
	RectangleType indictatorR;

	indictatorR.topLeft.x = itemR->topLeft.x;
	indictatorR.topLeft.y = itemR->topLeft.y + itemR->extent.y - 2;
	indictatorR.extent.x = itemR->extent.x;
	indictatorR.extent.y = 2;
				
	if (draw) {
		UInt16 i;
		UInt16 err;
        // CustomPatternType is 4 x 16 bits for OSVersion < 3.5, else it is
        // 8 x 8 bits.  What a pain.
        //
		// CustomPatternType pattern;
        UInt16 pattern[4];
		CustomPatternType savedPattern;

		WinGetPattern ((CustomPatternType *) &savedPattern);

		for (i = 0; i < 4; i++) {
			pattern[i] = 0xAA55;		
		}

		WinSetPattern ((const CustomPatternType *) &pattern);
	
		winH = WinSaveBits (&indictatorR, &err);
	
		WinFillRectangle (&indictatorR, 0);
	
		WinSetPattern ((const CustomPatternType *) &savedPattern);
	
	} else WinRestoreBits (savedBits, indictatorR.topLeft.x, indictatorR.topLeft.y);
	
	return (winH);
}

/**********************************************************
 * MainFormSelectTableItem
 * This routine either selects or unselects the specified
 * table item.				
 * While a table item is "on the move," having it be selected
 * can cause the Table code grief.  So, instead of using
 * the TblSelectItem()/TblUnhighlightSelect() calls, we now
 * manually select/unselect the table's row.  Before color,
 * only WinInvertRectangle() was called, so this is now in line
 * again with the way things used to work.  Sigh.
 **********************************************************/

static void MainFormSelectTableItem (Boolean selected, TablePtr table, Int16 row, Int16 column, RectangleType *r)
{
	// Get the item's rectangle.

	TblGetItemBounds (table, row, column, r);
	
	if (OSVersion < OS35)
		WinInvertRectangle (r, 0);
	else {

		// Set up the drawing state the way we want it.

		WinPushDrawState();
		WinSetBackColor(UIColorGetTableEntryIndex(UIFieldBackground));
		WinSetForeColor(UIColorGetTableEntryIndex(UIObjectForeground));
		WinSetTextColor(UIColorGetTableEntryIndex(UIObjectForeground));
		

		// Erase and (re)draw the item.

		WinEraseRectangle(r, 0);
		MainFormDrawTableItem(table, row, column, r);
	
		// If selected, make it look that way.

		if (selected)
			ReplaceTwoColors (r, 0,
						  TblGetItemInt(table, row, kRecordColumn), UIFieldBackground,
						  UIObjectSelectedForeground, UIObjectSelectedFill);

		// Restore the previous drawing state.

		WinPopDrawState();
	}
}

/***********************************************************************
 * ReplaceTwoColors:
 * This routine does a selection or deselection effect by
 * replacing foreground and background colors with a new pair
 * of colors. In order to reverse the process, you must pass
 * the colors in the opposite order, so that the current
 * and new colors are known to this routine. This routine
 * correctly handling the cases when two or more of these
 * four colors are the same, but it requires that the
 * affected area of the screen contains neither of the
 * given NEW colors, unless these colors are the same as
 * one of the old colors.
 ***********************************************************************/

static void ReplaceTwoColors (const RectangleType *rP, UInt16 cornerDiam,
							  UIColorTableEntries oldForeground, UIColorTableEntries oldBackground,
							  UIColorTableEntries newForeground, UIColorTableEntries newBackground)
{
	UInt8 oldForegroundIndex = UIColorGetTableEntryIndex(oldForeground);
	UInt8 oldBackgroundIndex = UIColorGetTableEntryIndex(oldBackground);
	UInt8 newForegroundIndex = UIColorGetTableEntryIndex(newForeground);
	UInt8 newBackgroundIndex = UIColorGetTableEntryIndex(newBackground);

	WinPushDrawState();
	WinSetDrawMode(winSwap);
	WinSetPatternType (blackPattern);

	if (newBackgroundIndex == oldForegroundIndex)
		if (newForegroundIndex == oldBackgroundIndex)
		{
			// Handle the case when foreground and background colors change places,
			// such as on black and white systems, with a single swap.
			WinSetBackColor(oldBackgroundIndex);
			WinSetForeColor(oldForegroundIndex);
			WinPaintRectangle(rP, cornerDiam);
		}
		else
		{
			// Handle the case when the old foreground and the new background
			// are the same, using two swaps.
			WinSetBackColor(oldForegroundIndex);
			WinSetForeColor(oldBackgroundIndex);
			WinPaintRectangle(rP, cornerDiam);
			WinSetBackColor(oldBackgroundIndex);
			WinSetForeColor(newForegroundIndex);
			WinPaintRectangle(rP, cornerDiam);
		}
	else if (oldBackgroundIndex == newForegroundIndex)
	{
		// Handle the case when the old background and the new foreground
		// are the same, using two swaps.
		WinSetBackColor(newForegroundIndex);
		WinSetForeColor(oldForegroundIndex);
		WinPaintRectangle(rP, cornerDiam);
		WinSetBackColor(newBackgroundIndex);
		WinSetForeColor(oldForegroundIndex);
		WinPaintRectangle(rP, cornerDiam);
	}
	else
	{
		// Handle the case when no two colors are the same, as is typically the case
		// on color systems, using two swaps.
		WinSetBackColor(oldBackgroundIndex);
		WinSetForeColor(newBackgroundIndex);
		WinPaintRectangle(rP, cornerDiam);
		WinSetBackColor(oldForegroundIndex);
		WinSetForeColor(newForegroundIndex);
		WinPaintRectangle(rP, cornerDiam);
	}

	WinPopDrawState();
}

/***************************************
 * MainFormDragMemo:
 * Performs the drag and drop operation
 * implementing Manual sorting.
 ***************************************
 */

static void MainFormDragMemo(EventPtr eventP)
{
	Int16 sortOrder;
	UInt16 selectedRow;
	UInt16 recordNum;
	UInt16 selectedRecord;
	UInt16 recAttributes;

	Int16 x, y;
	Boolean penDown = true;
	Boolean moving = false;
	Boolean selected = true;

	TablePtr tableP;
	Int16 row;
	UInt16 column;

	WinHandle savedBits;
	RectangleType rec;

	ListPtr listP;
	FormPtr formP;
	UInt32 time;
	UInt8 choice;
//USES_ALERT;
	formP = FrmGetActiveForm();

	sortOrder = GetMemoPadSortOrder(dbMemoPadRef);

	tableP = eventP->data.tblSelect.pTable;
	row = eventP->data.tblSelect.row;
	column = eventP->data.tblSelect.column;

	/* Highlight the selected record */

	selectedRecord = TblGetRowID(tableP, row);

	MainFormSelectTableItem(selected, tableP, row, column, &rec);

	/* Trace the pen until it moves enough to constitute
	 * a move operation or until the pen it released.
	 */

//	time = TimGetTicks();

	while (true) {

		PenGetPoint (&x, &y, &penDown);
		if (!penDown) break;

		if (!moving) {

			/* Draw a list of choices when user keep the pen
			 * for more than 2 seconds on a line with several
			 * proposal as beam, category, delete...
			 */
/*
			if ((time + 60) < TimGetTicks()) {
				listP = FrmGetObjectPtr(formP, FrmGetObjectIndex(formP, lstMainListChoices));

				// Position the list on the selected memo
				LstSetPosition (listP, x, y);
*/

//				choice = LstPopupList(listP);

/*
				LstSetSelection (listP, -1);

				// If user choosed edit, let the function handle the edition.

				if (MainFormHandleListChoices (choice, selectedRecord) == true) {
					if (choice == 3) return;

					// Unhighlight the item and exit the function.

					if (TblFindRowID(tableP, selectedRecord, &selectedRow))
						MainFormSelectTableItem(false, tableP, selectedRow, column, &rec);
					return;
				}
			}
*/
			if (sortOrder != soAlphabetic) {

				/* Is the pen still within the bounds of the selected item?
				 * if not, draw the move indicator.
				 */

				if (!RctPtInRectangle (x, y, &rec)) {
					moving = true;
					TblGetItemBounds (tableP, row, column, &rec);
					savedBits = MainFormInvertMoveIndicator (&rec, 0, true);
				}

			} else selected = RctPtInRectangle(x, y, &rec);

		} else {
//			time = TimGetTicks();

			if (!RctPtInRectangle (x, y, &rec)) {

				/* Above the first item? */
				if (row < 0) {
					if (y >= rec.topLeft.y) {
						row++;
						MainFormInvertMoveIndicator(&rec, savedBits, false);
						rec.topLeft.y -= rec.extent.y;
						savedBits = MainFormInvertMoveIndicator (&rec, 0, true);
					}					
				}

				/* Move up */
				else if (y < rec.topLeft.y) {
					recordNum = TblGetRowID(tableP, row);

					if (MainFormSeekRecord (&recordNum, 1, dmSeekBackward))
						{
						MainFormInvertMoveIndicator(&rec, savedBits, false);
						if (row) row--;
						else {				
							MainFormScrollLines(-1);
							if (TblFindRowID(tableP, selectedRecord, &selectedRow))
								MainFormSelectTableItem(selected, tableP, selectedRow, column, &rec);
						} 
						TblGetItemBounds (tableP, row, column, &rec);
						savedBits = MainFormInvertMoveIndicator (&rec, 0, true);

					} else if (row == 0) {
						row--;
						MainFormInvertMoveIndicator (&rec, savedBits, false);
						rec.topLeft.y -= rec.extent.y;
						savedBits = MainFormInvertMoveIndicator (&rec, 0, true);
					}
				}

				/* Move down */
				else {
					recordNum = TblGetRowID(tableP, row);

					if (MainFormSeekRecord(&recordNum, 1, dmSeekForward))
						{
						MainFormInvertMoveIndicator (&rec, savedBits, false);				
						if (row < TblGetLastUsableRow (tableP)) row++;
						else {
							MainFormScrollLines(1);
							if (TblFindRowID(tableP, selectedRecord, &selectedRow))
								MainFormSelectTableItem(selected, tableP, selectedRow, column, &rec);
						}

						TblGetItemBounds (tableP, row, column, &rec);
						savedBits = MainFormInvertMoveIndicator (&rec, 0, true);
					}
				}
			}
		}
	}

	/* Turn off the move indicator, if it is on */

	if (moving) savedBits = MainFormInvertMoveIndicator (&rec, savedBits, false);

	/* If the highlighted item is visible, unhighlight it */
		if (TblFindRowID(tableP, selectedRecord, &selectedRow))
			MainFormSelectTableItem(false, tableP, selectedRow, column, &rec);

	if (moving) {
		if (row >= 0) {
			recordNum = TblGetRowID (tableP, row);

			if (selectedRecord == recordNum) {
				return;
            }
			
			recordNum++;

		} else recordNum = TblGetRowID(tableP, 0);

		DmMoveRecord (dbCryptoPadRef, selectedRecord, recordNum);

		MainFormLoadRecords();
		TblRedrawTable (tableP);
        
        // If to is > from in DmMoveRecord, the record ends up at to - 1
        if (recordNum > selectedRecord) {
            recordNum--;
        }
		if (TblFindRowID(tableP, recordNum, &selectedRow)) {
            currentRow = selectedRow;
        }
        else {
            currentRow = 0;
        }
	}
		
	/* If we didn't move the item then it's been selected for editing,
	 * go to the Edit field.
	 */

	else if (sortOrder != soAlphabetic || selected) {
        MainFormEditRecord(tableP, row);
	}
}

/******************************************
 * MainFormEditRecord:
 * v4.11: carved out code for entering a
 * record for edit into separate function
 * so it could be conveniently used in
 * 5-way nav code.
 ******************************************
 */
static void MainFormEditRecord(TablePtr tableP, Int16 row)
{
	UInt16 recAttributes;

	currentRecord = TblGetRowID(tableP, row);

	editScrollPosition = 0;

	// Get the category and secret attribute of the current record.

	DmRecordInfo (dbCryptoPadRef, currentRecord, &recAttributes, NULL, NULL);

	// If this is a "private" record, then determine what is to be shown.

	if (recAttributes & dmRecAttrSecret) {
		switch (PrivateRecordVisualStatus) {
			case showPrivateRecords:
				MainFormEnterTableItem();
				break;
					
			case maskPrivateRecords:
				if (SecVerifyPW (showPrivateRecords) == true) {
				//	Make it work as PalmOS 4, that is mask again the memo when edited
				//PrivateRecordVisualStatus = showPrivateRecords;
				MainFormEnterTableItem();
				}
				break;
		}
	} else MainFormEnterTableItem();
}

/******************************************
 * MainFormEnterTableItem:
 * Handle the choice of the list that
 * is displayed when user keep the pen
 * for more than 2 seconds on a line with several
 * proposal as beam, category, delete...
 ******************************************
 */

Boolean MainFormHandleListChoices(UInt8 choice, UInt16 recordNum)

{
	UInt16 index;

	MemPtr		p;
	MemPtr		newRecP;
	CryptoRecordPtr cryptoRecordP;
UInt16 offset;
	MemHandle handle, newHandle;
	MemoRecordType memoRecord;
	MemoRecordPtr memoRecordP;
MemoRecordPtr srcMemoRecordP;
	UInt16 newIndex;	
	UInt16 recAttributes;
	UInt16 nLength;			
	UInt32 recordID;

USES_ALERT;
	switch (choice) {

		// Edit
		case 0:
		return false;
		break;

		// Beam
		case 1:
		if (!isEncrypted(recordNum, dbCryptoPadRef))
			MemoSendRecord (dbMemoPadRef, recordNum);
		return true;
		break;

		// Delete
		case 2:

		if (!isEncrypted(recordNum, dbCryptoPadRef))	{
	
			UInt16 btn;
			FormPtr alert;
			Boolean backup;
			UInt16 ctlIndex;

			index = GetMemoPadRecordIndex(recordNum);
			alert = FrmInitForm (frmDeleteForm);
	
			ctlIndex = FrmGetObjectIndex (alert, chkDeleteBackup);
			FrmSetControlValue (alert, ctlIndex, saveBackup);
			btn = FrmDoDialog (alert);
			backup = FrmGetControlValue (alert, ctlIndex);
	
			FrmDeleteForm (alert);

			if (btn  == btnDeleteCancel)
				return (true);
	
			saveBackup = backup;
		} else {
			if (FrmAlert(altConfirmDeleteAlert) == 1) return (true);
		}

		// Delete or archive the record.
		if (!isEncrypted(recordNum, dbCryptoPadRef)) {
			if (saveBackup)
				DmArchiveRecord (dbMemoPadRef, index);
			else
				DmDeleteRecord (dbMemoPadRef, index);
			DmMoveRecord (dbMemoPadRef, index, DmNumRecords (dbMemoPadRef));
		}
 
		DmRemoveRecord (dbCryptoPadRef, recordNum);

		memosInCategory--;
		currentRecord = noRecordSelected;
		
		FrmUpdateForm(frmMainForm, updateTable);
		return true;
		break;

		// Duplicate
		case 3:

if (isEncrypted(recordNum, dbCryptoPadRef))	{
	index = dmMaxRecordIndex;

	/* Create a new record in the MemoPad database */
	handle = DmQueryRecord(dbCryptoPadRef, recordNum);
	cryptoRecordP = MemHandleLock(handle);

		nLength = sizeof(cryptoRecordP->flag) +
				sizeof(cryptoRecordP->passwdLen) +
				sizeof(cryptoRecordP->textLen) +
				cryptoRecordP->passwdLen + 1 +
				cryptoRecordP->textLen + 1 +
				sizeof(cryptoRecordP->tableFont) +
				sizeof(cryptoRecordP->fieldFont) +
				sizeof(cryptoRecordP->tableColor) +
				sizeof(cryptoRecordP->fieldColor);

	newHandle = DmNewRecord(dbCryptoPadRef, &index,nLength);

	if (!newHandle) {
		FrmAlert(DeviceFullAlert);
		return (true);
	}

	/* Lock down the block containing the new record */

	p = MemHandleLock(newHandle);

	/* Write the field's text to the database record */

	DmWrite(p, 0, cryptoRecordP, nLength);

	/* Unlock the block of the new record */

	MemPtrUnlock(p);
	MemPtrUnlock(cryptoRecordP);

	/* Set the category of the record */
	DmRecordInfo(dbCryptoPadRef, recordNum, &recAttributes, 0, 0);

	DmReleaseRecord(dbCryptoPadRef, index, true);

	DmSetRecordInfo(dbCryptoPadRef, index, &recAttributes, 0);
} else {

	/* Create a new record in the MemoPad database */
	index = GetMemoPadRecordIndex(recordNum);
	handle = DmQueryRecord(dbMemoPadRef, index);
	p = MemHandleLock(handle);

	nLength = StrLen(p);

	newIndex = dmMaxRecordIndex;
	newHandle = DmNewRecord(dbMemoPadRef, &newIndex,nLength + 1);

	if (!newHandle) {
		FrmAlert(DeviceFullAlert);
		return (true);
	}

	/* Lock down the block containing the new record */

	newRecP = MemHandleLock(newHandle);

	/* Write the field's text to the database record */

	DmWrite(newRecP, 0, p, nLength + 1);

	/* Unlock the block of the new record */

//	MemPtrUnlock(newRecP);
	MemHandleUnlock(newHandle);

//	MemPtrUnlock(p);
	MemHandleUnlock(handle);

	/* Set the category of the record */
	DmRecordInfo(dbMemoPadRef, index, &recAttributes, 0, 0);

	DmReleaseRecord(dbMemoPadRef, newIndex, true);

	DmSetRecordInfo(dbMemoPadRef, newIndex, &recAttributes, 0);
	
	DmRecordInfo(dbMemoPadRef, newIndex, 0, &recordID, 0);

		handle = DmQueryRecord(dbCryptoPadRef, recordNum);
	srcMemoRecordP = MemHandleLock(handle);

newIndex = dmMaxRecordIndex;
newHandle = DmNewRecord(dbCryptoPadRef, &newIndex, sizeof(MemoRecordType));
			if (newHandle == 0) {
				FrmAlert(DeviceFullAlert);
				return;
			}

	/* Lock down the block containing the new record */

memoRecordP = MemHandleLock(newHandle);

	/* Write the field's text to the database record */

	DmWrite(memoRecordP, 0, srcMemoRecordP,  sizeof(MemoRecordType));

	memoRecord.flag = recDecrypted;
	memoRecord.uniqueID = recordID;

	DmWrite(memoRecordP, 0, &memoRecord,
							sizeof(memoRecord.uniqueID)+sizeof(memoRecord.flag)+1);


	/* Unlock the block of the new record */

//	MemPtrUnlock(memoRecordP);
MemHandleUnlock(newHandle);
MemHandleUnlock(handle);
//	MemPtrUnlock(srcMemoRecordP);

	DmRecordInfo(dbCryptoPadRef, recordNum, &recAttributes, 0, 0);

	DmReleaseRecord(dbCryptoPadRef, newIndex, true);

	DmSetRecordInfo(dbCryptoPadRef, newIndex, &recAttributes, 0);
}
memosInCategory++;
currentRecord = noRecordSelected;
FrmUpdateForm(frmMainForm, updateTable);
		return true;
		break;

		// Category
		case 4:

		return true;
		break;

		// Details
		case 5:
		currentRecord = recordNum;
		FrmPopupForm(frmDetailsForm);
//		FrmUpdateForm(frmMainForm, updateTable);
		return true;
		break;

		default:
		return true;
	}
}

/******************************************
 * MainFormEnterTableItem:
 * A item in the table of the main form
 * has been selected.
 ******************************************
 */

void MainFormEnterTableItem(void)

{
	if (isEncrypted(currentRecord, dbCryptoPadRef)) {
		EventType event;

		if ((prefsGlobalPasswordP != 0) && (globalPasswordP == 0)) {
			event.eType = globalPasswordStatusEvent;
			event.data.generic.datum[1] = queryGlobalPassword;
			event.data.generic.datum[2] = recEncrypted;

			FrmPopupForm(frmQueryGlobalPasswordForm);
			EvtAddEventToQueue(&event);
			return;
		}

		/* The record is encrypted.
		 * Pops the password form up.
		 */

		event.eType = memoStatusEvent;
		event.data.generic.datum[1] = recEncrypted;

		FrmPopupForm(frmPasswordForm);
		EvtAddEventToQueue(&event);
	} else FrmGotoForm(frmEditForm);
}

/******************************************
 * MainFormInit:
 * Initialize the main form and sets up the
 * table drawing callback functions.
 ******************************************
 */

static void MainFormInit(void)
{
	ControlPtr triggerP;
	UInt16 category;

	FormPtr formP;
	UInt16 fieldIndex;
	FieldPtr fieldP;

/*
		Char				title[timeStringLength];
	DateTimeType 	dateTime;
	
	titleTemplateH = DmGetResource ('tSTR', strEditTitleMemo);
	titleTemplateP = MemHandleLock (titleTemplateH);
	length = StrLen (timeStringLength) + StrLen (titleTemplateP);
StrNCopy (titleP, titleTemplateP, length);

	TimSecondsToDateTime (TimGetSeconds (), &dateTime);
	TimeToAscii (dateTime.hour, dateTime.minute, TimeFormat, title);
	*/
//	FrmCopyTitle (FrmGetActiveForm (), title);

/*
	StrIToA (posStr, pos+1);
	StrIToA (totalStr, memosInCategory);

*/

	/* Now get space for the new title */

/*

	titleP = MemPtrNew (length);
	ErrFatalDisplayIf (!titleP, "Out of memory");


	strP = StrChr (titleTemplateP, '#');
	ErrNonFatalDisplayIf (!strP, "Missing '#' in MemoEdit'tStr' strEditTitleMemo");

	length = strP - titleTemplateP;
	
	titleP[length] = 0;
	StrCat (titleP, posStr);

	strP++;
	StrCat (titleP, strP);
	StrCat (titleP, totalStr);

	FrmSetTitle (frm, titleP);
*/
	/* Show and set the focus to 
	 *the Look Up field if the preferences
	 are so. */

	MainFormDisplayLookupField();


	triggerP = GetActiveFormObjectPtr(popMainCategory);

	if (showAllCategories)
		category = dmAllCategories;
	else 
		category = currentCategory;

	/* Loads the record in the table */

	MainFormLoadRecords();

	/* Updates the currently selected category */

	CategoryGetName(dbMemoPadRef, category, categoryName);
	CategorySetTriggerLabel(triggerP, categoryName);

	currentForm = frmMainForm;
}

/****************************************
 * DrawStringWithEllipsis:
 * This routine draws the title of the specified memo. 
 ****************************************
 */

void DrawStringWithEllipsis (Char * textP, Int16 x, 
			Int16 y, Int16 width, JustificationEnum just)
{

// ***************** Code for PalmOS 3.0 to 3.3 *************//
	Boolean fits;
	Char ellipsisCharP[2];
	UInt16		strWidth = width;
	UInt16		maxStrWidth = width;
	UInt16		ellipsWidth = 0;
	UInt16 textLength;

	fits = FontCharsInWidth(textP, &strWidth, &textLength);

	// If the string fits, draw it entirely
	if (!fits) {
		MemSet(ellipsisCharP , 2, 0);
		ChrHorizEllipsis (&(ellipsisCharP[0]));

		// Take out the width of an ellipsis char
		ellipsWidth = FntCharWidth ((ellipsisCharP[0]));
		maxStrWidth -= ellipsWidth;

		// Keep removing last char until name fits
		while ((textLength > 0) && (strWidth > maxStrWidth)) {
			textLength--;
			strWidth -= FntCharWidth (textP[textLength]);
		}
	}

	if (just == justCenter)
		x += ((width - strWidth - ellipsWidth) / 2);

	// Draw the string
	WinDrawChars(textP, textLength, x, y);

	// Draw the ellipsis, if necessary
	if (ellipsWidth != 0) {
		x += strWidth;
		WinDrawChars (ellipsisCharP, 1, x, y);
	}
// ***************** End of code for PalmOS 3.0 to 3.2 *************//
/*
	Char * ptr = StrChr (textP, linefeedChr);
	Int16 strWidth;

	UInt16 titleLen = (ptr == NULL ? StrLen (textP) : (UInt16) (ptr - textP));

	// If the string fits, draw it entirely
	if (FntWidthToOffset (textP, titleLen, width, NULL, &strWidth) == titleLen) {
		if (just == justCenter)
			x += ((width - strWidth) / 2);

		WinDrawChars (textP, titleLen, x, y);
	} else {
		Int16 ellipsisWidth;

		ellipsisWidth = FntCharWidth (chrEllipsis);
		titleLen = FntWidthToOffset (textP, titleLen, width - ellipsisWidth, NULL, &strWidth);
		if (just == justCenter)
			x += ((width - strWidth - ellipsisWidth) / 2);

		WinDrawChars (textP, titleLen, x, y);

		// Draw the ellipsis
		WinDrawChar (chrEllipsis, x + strWidth, y);
	}
*/
}


/*******************************************
 * DrawMemoPadRecordText:
 * Draw each usable table cell
 * that contains a MemoPad record.
 *******************************************
 */

static void DrawMemoPadRecordText(UInt16 recordNum, UInt16 x, RectanglePtr bounds)
{
	MemHandle handle;
	Char * textP;

	Int16 cellWidth;
	UInt16 index;

	/* Get the index of the MemoPad record */

	index = GetMemoPadRecordIndex(recordNum);

	handle = DmQueryRecord(dbMemoPadRef, index);
	textP = MemHandleLock(handle);

	// Determine the length of text that will fit within the bounds.

	cellWidth = bounds->extent.x - x;

	DrawStringWithEllipsis (textP, x, bounds->topLeft.y, cellWidth, justLeft);
	
	// Unlock the handle to the record.

	MemHandleUnlock(handle);
}


 /*******************************************
 * DrawCryptoPadRecordText:
 * Draw each usable table cell
 * that contains a CryptoPad record.
 *******************************************
 */

 static void DrawCryptoPadRecordText(UInt16 recordNum, UInt16 x, RectanglePtr bounds)
{
	MemHandle handle;
	CryptoPackedRecordPtr packedP;
	CryptoRecordType record;

	UInt16 right;
	Int16 cellWidth;

	/* Adjust spacing with magic values! */
	right = 7;

	handle = DmGetRecord(dbCryptoPadRef, recordNum);
	packedP = MemHandleLock(handle);
	UnpackRecord(&record, packedP);

	// Determine the length of text that will fit within the bounds.
	
	cellWidth = bounds->extent.x - x - right;

	DrawStringWithEllipsis (record.textP, x, bounds->topLeft.y, cellWidth, justLeft);

	MemHandleUnlock(handle);
	DmReleaseRecord(dbCryptoPadRef, recordNum, false);

	/* Draw the little icon */

	if (isEncrypted(recordNum, dbCryptoPadRef)) {
		BitmapPtr bitmapP;

		handle = DmGetResource('Tbmp', bmpMainKeyBitmap);
		bitmapP = MemHandleLock(handle);
		WinDrawBitmap(bitmapP, bounds->extent.x - right, bounds->topLeft.y);
		MemHandleUnlock(handle);
		DmReleaseResource(handle);
	}
}
 
 /*******************************************
 * MainFormDrawTableItem:
 * Callback to draw each usable table cell
 *******************************************
 */

static void MainFormDrawTableItem(void * tableP,
				Int16 row, Int16 column, RectanglePtr bounds)
{
	UInt16 recordNum;
	UInt16 posInCategory;

	char posStr [6];
	UInt16 length;
	UInt16 x, y;
	UInt16 category;
	UInt16 recAttributes;

	/* Draw the record number */

	recordNum = TblGetRowID(tableP, row);

	if (showAllCategories) category = dmAllCategories;
	else category = currentCategory;

	posInCategory = DmPositionInCategory(dbCryptoPadRef, recordNum, category);

	x = bounds->topLeft.x + 1;
	y = bounds->topLeft.y;

	if (useSingleFont == separateFont) {
		// Font.
		FntSetFont(TblGetItemFont (tableP, row, kRecordColumn));
		// Color
		if (OSVersion >= OS35) WinSetTextColor(TblGetItemInt(tableP, row, kRecordColumn));
	} else {
		// Font
		FntSetFont(currentTableFont);
		// Color
		if (OSVersion >= OS35) WinSetTextColor(currentTableColor);
	}

	/* Format the memo's postion, within its category, 
	* an draw it.
	*/

	StrIToA (posStr, posInCategory+1);
	length = StrLen(posStr);
	ErrNonFatalDisplayIf(length > sizeof(posStr) - 2, "Too many records");
	posStr[length++] = '.';
	posStr[length] = 0;
	
	if (length < 3) x += FntCharWidth ('1');
	WinDrawChars (posStr, length, x, y);
	x += FntCharsWidth (posStr, length) + 4;	

	// If we are here then we either mask the memo out or display the
	// memo title.

	DmRecordInfo (dbCryptoPadRef, recordNum, &recAttributes, NULL, NULL);
   	if (((recAttributes & dmRecAttrSecret) 
   	&& PrivateRecordVisualStatus == maskPrivateRecords)) {
		RectangleType maskRectangle;

		MemMove (&maskRectangle, bounds, sizeof (RectangleType));
		maskRectangle.topLeft.x = x;
		maskRectangle.extent.x = bounds->extent.x - x;
		MainFormDisplayMask (&maskRectangle);

	} else {

		/* Draw the record text */

		if (isEncrypted(recordNum, dbCryptoPadRef)) {
			DrawCryptoPadRecordText(recordNum, x, bounds);
		} else DrawMemoPadRecordText(recordNum, x, bounds);
	}
}

/*******************************************
 * MainFormDisplayMask
 * Draws the masked display for the record.
  ******************************************/
static void MainFormDisplayMask (RectanglePtr bounds)
{
	RectangleType tempRect;
	CustomPatternType origPattern;
	MemHandle	bitmapH;
	BitmapType * bitmapP;

	CustomPatternType pattern = {0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55};

	MemMove (&tempRect, bounds, sizeof (RectangleType));

	// Make sure it fits nicely into the display.

	tempRect.topLeft.y++;
	tempRect.extent.y --;
	tempRect.extent.x -= SecLockWidth + 1;

	WinGetPattern((CustomPatternType *) &origPattern);
	WinSetPattern ((const CustomPatternType *) &pattern);
	WinFillRectangle (&tempRect, 0);
	WinSetPattern((const CustomPatternType *) &origPattern);
	
	/* draw the lock icon */

	bitmapH = DmGetResource (bitmapRsc, SecLockBitmap);
	if (bitmapH) {
		bitmapP = MemHandleLock (bitmapH);
		WinDrawBitmap (bitmapP, tempRect.topLeft.x + tempRect.extent.x + 1 , 
			tempRect.topLeft.y + ((tempRect.extent.y - SecLockHeight) / 2));
		MemPtrUnlock (bitmapP);
	}
}

/********************************************************
 * MainFormNumberOfRows
 * This routine return the maximun number of visible rows,
 * with the current list view font setting.
 ********************************************************/
static Int16 MainFormNumberOfRows (TablePtr tableP)
{
	Int16	rowsInTable;
	Int16 nRows = 0;
	UInt16 tableHeight;
	RectangleType	r;

	rowsInTable = TblGetNumberOfRows (tableP);

	TblGetBounds (tableP, &r);
	tableHeight = r.extent.y;

	if (useSingleFont == separateFont) {
		UInt16	recordNum;
		MemHandle nextRecord;
		UInt16 dataHeight = 0;
	
		recordNum = topRecord;

		while (true) {
			nextRecord = DmQueryNextInCategory (dbCryptoPadRef, &recordNum, currentCategory);

			if ((dataHeight > tableHeight)) {
				nRows--;
				break;
			}
			if (!nextRecord) break;

			GetRecordFont(recordNum);
			FntSetFont (currentTableFont);
			dataHeight += FntLineHeight ();
			nRows++;
			recordNum++;
		}
	} else {
		FntSetFont (currentTableFont);
		nRows = tableHeight / FntLineHeight ();
	}

	if (nRows <= rowsInTable) 
		return (nRows);
	else
		return (rowsInTable);
}


/******************************************
 * MainFormLoadTable:
 * Load the table with the records.
 ******************************************
 */

static void MainFormLoadTable(void)
{	
	TablePtr tableP;
	Int16 recordNum;
	Int16 numRows;

	MemHandle record;
	Int16 nRow;
	UInt16 lineHeight;
	UInt16 dataHeight;
	UInt16 tableHeight;
	RectangleType	r;

	tableP = GetActiveFormObjectPtr(tblMainRecordList);
	numRows = TblGetNumberOfRows(tableP);

	/* Set the correct topRecord and topDB
	 * according to the category.
	 */

	recordNum = topRecord;

	TblGetBounds (tableP, &r);
	tableHeight = r.extent.y;
	dataHeight = 0;

	/* For each row in the table, store the 
	* record number in the table item
	* that will display the record.
	*/

	for (nRow = 0; nRow < numRows; nRow++) {

		record = DmQueryNextInCategory (dbCryptoPadRef, &recordNum, currentCategory);

		if (record && (useSingleFont == separateFont)) { 
			GetRecordFont(recordNum);
			FntSetFont (currentTableFont);
		}

		lineHeight = FntLineHeight ();

		if (record && (tableHeight >= dataHeight + lineHeight)) {
			TblSetItemStyle(tableP, nRow, kRecordColumn, customTableItem);
			TblSetItemFont (tableP, nRow, kRecordColumn, currentTableFont);
			TblSetRowHeight (tableP, nRow, lineHeight);
			TblSetRowID(tableP, nRow, recordNum);
			TblSetItemInt(tableP, nRow, kRecordColumn, currentTableColor);
			TblSetRowUsable(tableP, nRow, true);

		/* Marks the row invalid so that is is redrawn */

		TblMarkRowInvalid(tableP, nRow);
				
		} else {

			// Set the row height - when scrolling down, the heights of the last rows of
			// the table are used to determine how far to scroll.  As rows are deleted
			// from the top of the table, formerly unused rows scroll into view, and the
			// height is used before the next call to ListViewLoadTable (which would set
			// the height correctly).

			TblSetRowHeight (tableP, nRow, lineHeight);
			TblSetRowUsable(tableP, nRow, false);
		}

	if (nRow+1 < numRows) recordNum++;

	dataHeight += lineHeight;
	}
	MainFormUpdateScrollBar();
}

/******************************************
 * MainFormLoadRecords:
 * Load the records that will be displayed
 * in the table.
 ******************************************
 */

static void MainFormLoadRecords(void)
{
	TablePtr tableP;

	MemHandle record;

	if (showAllCategories) currentCategory = dmAllCategories;

	tableP = GetActiveFormObjectPtr(tblMainRecordList);

	/* Try to show a full display of records 
	 *  and perform positioning on the table.
	 */

	MainFormShowFullDisplay();

	/* Load each row of the table with a record.
	 * If no record could be found, mark the row
	 * unusable.
	 */

	MainFormLoadTable();

	/* Set the table's custom draw procedure */

	TblSetCustomDrawProcedure(tableP, kRecordColumn, MainFormDrawTableItem);

	/* Mark the columns as usable otherwise
	 * the table won't be displayed by default.
	 */

	TblSetColumnUsable(tableP, kRecordColumn, true);

	/* Updates the scrollbars */

	MainFormUpdateScrollBar();
}

/******************************************
 * MainFormShowFullDisplay:
 * Show a full diplay in the main form.
 ******************************************
 */

static void MainFormShowFullDisplay(void)
{
	UInt16 recordNum;
	MemHandle previousRecord;
	TablePtr tableP;
	Int16 nRow, numRows = 0;
	UInt32 recordID;

	UInt16 dataHeight = 0;
	UInt16 tableHeight;
	RectangleType	r;

	tableP = GetActiveFormObjectPtr(tblMainRecordList);
	numRows = MainFormNumberOfRows (tableP);

	TblGetBounds (tableP, &r);
	tableHeight = r.extent.y;


	if (currentRecord != noRecordSelected) {
	/* There is a current record. Make it visible.
	 * If the current record index is smaller than the top index
	 * of table, update the top index.
	 * Otherwise, find the record displayed at the bottom of the table
	 * If the record is before the current record then the 
	 * current record isn't displayed.  In that case make
	 * the current record the first record displayed in the table.
	 */

		// Is the current record before the first visible record?
		if (topRecord > currentRecord)
			topRecord = currentRecord;
		else {
		// Is the current record after the last visible record?
			recordNum = topRecord;

			DmSeekRecordInCategory (dbCryptoPadRef, &recordNum, numRows-1,
					dmSeekForward, currentCategory);
			if (recordNum < currentRecord) topRecord = currentRecord;
		}
	}
	

	/* Try to show a full display of records.  Starting at the last
	 * record and working backwards, find the record displayed at the 
	 * top of the table.  If the record is before the topRecord
	 * then topRecord is set too far down the list of records.
	 * Set topRecord to the record one screen full from the end.
	 */

	if (memosInCategory) {
		if (useSingleFont == separateFont) {
			Err find = 1;
			UInt8 offset = 0;

			recordNum = dmMaxRecordIndex;


			while ((dataHeight <= tableHeight) && find) {
				find = MainFormSeekRecord (&recordNum, offset, dmSeekBackward);

				GetRecordFont(recordNum);

				FntSetFont (currentTableFont);
				dataHeight += FntLineHeight ();
				offset = 1;
			}

			if (find) MainFormSeekRecord (&recordNum, 1, dmSeekForward);

			topRecord = min (topRecord, recordNum);
		
		} else {
			recordNum = dmMaxRecordIndex;
			MainFormSeekRecord (&recordNum, (numRows-1), dmSeekBackward);
			topRecord = min (topRecord, recordNum);
		}
	} else topRecord = 0;
}

/******************************************
 * DecryptRecord:
 * Decrypt the current record.
 ******************************************
 */

void DecryptRecord(UInt16 recordNum, Char * passwordP)
{
	MemHandle handle;
	CryptoPackedRecordPtr packedP;
	CryptoRecordType record;

	UInt16 nOffset;

	handle = DmGetRecord(dbCryptoPadRef, recordNum);
	packedP = MemHandleLock(handle);
	UnpackRecord(&record, packedP);

	/* Finds the offset to start decryption */

	nOffset = StrChr (record.textP, linefeedChr) - record.textP + 1;

	if (nOffset) {
		CryptoRecordType newRecord;

		/* Set the flag to reflect the status of the record */

		newRecord.flag      = recDecrypted;
		newRecord.passwdLen = 0;
		newRecord.textLen   = record.textLen;

		/* Get rid of the stored password */

		newRecord.passwordP = MemPtrNew(1);
		newRecord.passwordP[0] = '\0';

		/* Store the decrypted text in the record */

		newRecord.textP = MemPtrNew(record.textLen + 1);
		StrNCopy(newRecord.textP, record.textP, nOffset);
		Blowfish(record.textP + nOffset, newRecord.textP + nOffset,
					passwordP, newRecord.textLen - nOffset,
					BF_DECRYPT);

		/* Set the font settings */
		newRecord.tableFont = record.tableFont;
		newRecord.fieldFont = record.fieldFont;
		newRecord.tableColor = record.tableColor;
		newRecord.fieldColor = record.fieldColor;

		/* Replaces the record with the new one */

		MemHandleUnlock(handle);
		PackRecord(&newRecord, handle);

		/* Deallocates memory */

		MemPtrFree(newRecord.passwordP);
		MemPtrFree(newRecord.textP);
	}

	/* Release the record to the data manager */

	DmReleaseRecord(dbCryptoPadRef, recordNum, true);
}

/******************************************
 * EncryptCategory:
 * Encrypt the current category.
 ******************************************
 */

static void EncryptCategory()
{
	MemHandle memoPadHandle, cryptoPadHandle;

	UInt16 i, numRecInCategory;
	Int16 sortOrder;

	Char * textP;
	Boolean validate;

	FormPtr formP;
    ProgressInfoPtr progrInfoP;
    ProgrType progr;
	UInt16 progrinc;

	Char * strProgressP;
	MemHandle strProgressH;


	if (DmNumRecordsInCategory(dbMemoPadRef, currentCategory) == 0) {
		FrmAlert(altCategoryAlreadyEncryptedAlert);
		RemoveCurrentPassword();
		return;
	}

	numRecInCategory = DmNumRecordsInCategory(dbCryptoPadRef, currentCategory);

	formP = FrmInitForm(frmProgressForm);

	// Display the title of the progress form.

	strProgressH = DmGetResource('tSTR', strProgressEncryption);
	strProgressP = MemHandleLock(strProgressH);
		
	FrmSetTitle (formP, strProgressP);

	MemHandleUnlock(strProgressH);   
	DmReleaseResource(strProgressH); 

	FrmDrawForm(formP);

	currentRecord = 0;


	progrInfoP = ProgressFormInitAndDraw (formP);

	for (i = 0; i < numRecInCategory; i++) {
		progr.major.soFar = 0;
		progr.major.total = numRecInCategory;
		progr.minor.soFar = 0;
		progr.minor.total = 0;
		progrinc = i ;
		progr.major.soFar = progrinc;

		DmSeekRecordInCategory (dbCryptoPadRef, &currentRecord, 0, dmSeekForward, currentCategory);


		if (!isEncrypted(currentRecord, dbCryptoPadRef)) {

			memoPadHandle = DmQueryRecord(dbMemoPadRef, GetMemoPadRecordIndex(currentRecord));
			textP = MemHandleLock(memoPadHandle);
	
			/* Validate the format of the record
			 * in checking if there is 2 lines.
			 */

			validate = (StrChr (textP, linefeedChr) ? true : false);

			progrInfoP->statusFunc (progrInfoP, &progr, textP);

			if (validate) {
				MemoRecordPtr memoRecordP;
				FontID tableFont, fieldFont;
				IndexedColorType tableColor, fieldColor;

				/* Get the font settings */

				if (useSingleFont == separateFont) {
					cryptoPadHandle = DmQueryRecord(dbCryptoPadRef, currentRecord);
					memoRecordP = MemHandleLock(cryptoPadHandle);

					tableFont = memoRecordP->tableFont;
					fieldFont = memoRecordP->fieldFont;
					tableColor = memoRecordP->tableColor;
					fieldColor = memoRecordP->fieldColor;

					MemHandleUnlock(cryptoPadHandle);
				}

				/* Change the structure of the CryptoPad record, 
				 * encrypt it and delete the record in MemoPad.
				 */

				ChangeRecordFromMemoPadToCryptoPad(textP, currentRecord, 
					tableFont, fieldFont, tableColor, fieldColor);

			} else MemHandleUnlock(memoPadHandle);

		}
		currentRecord++;
	}

	progr.major.soFar = progr.major.total;
	progrInfoP->statusFunc (progrInfoP,&progr,NULL);

	RemoveCurrentPassword();

	sortOrder = GetMemoPadSortOrder(dbMemoPadRef);
	if (sortOrder == soAlphabetic) {
		DmQuickSort(dbCryptoPadRef, 
			(DmComparF*) &CompareCryptoPadRecords, (Int16) sortOrder);
		DmQuickSort(dbMemoPadRef, 
			(DmComparF*) &CompareMemoPadRecords, (Int16) sortOrder);
	}

	currentRecord = noRecordSelected;

	FrmDeleteForm(formP);
	FrmGotoForm(frmMainForm);
}

/******************************************
 * DecryptCategory:
 * Decrypt the current category.
 ******************************************
 */

static void DecryptCategory()
{
	UInt16 i, numRecInCategory;
	Int16 sortOrder;

	FormPtr formP;
    ProgressInfoPtr progrInfoP;
    ProgrType progr;
	UInt16 progrinc;

	Char * strProgressP;
	MemHandle strProgressH;


	numRecInCategory = DmNumRecordsInCategory(dbCryptoPadRef, currentCategory);
	if (numRecInCategory == 0) {
		FrmAlert(altCategoryAlreadyDecryptedAlert);
		RemoveCurrentPassword();
		return;
	}

	formP = FrmInitForm(frmProgressForm);

	// Display the title of the progress form.

	strProgressH = DmGetResource('tSTR', strProgressDecryption);
	strProgressP = MemHandleLock(strProgressH);
		
	FrmSetTitle (formP, strProgressP);

	MemHandleUnlock(strProgressH);   
	DmReleaseResource(strProgressH); 


	FrmDrawForm(formP);


	currentRecord = 0;

	progrInfoP = ProgressFormInitAndDraw (formP);

	for (i = 0; i < numRecInCategory; i++) {

		progr.major.soFar = 0;
		progr.major.total = numRecInCategory;
		progr.minor.soFar = 0;
		progr.minor.total = 0;
		progrinc = i ;
		progr.major.soFar = progrinc;

		progrInfoP->statusFunc (progrInfoP,&progr,NULL);

		/* Seek the item after the currentRecord index.*/

		DmSeekRecordInCategory (dbCryptoPadRef, &currentRecord, 0, dmSeekForward, currentCategory);

		if (isEncrypted(currentRecord, dbCryptoPadRef)) {
			if (ValidatePassword(currentPasswordP)) {
				MemHandle cryptoPadHandle;
				FontID tableFont, fieldFont;
				IndexedColorType tableColor, fieldColor;

				CryptoPackedRecordPtr packedP;
				CryptoRecordType cryptoRecord;
			
				DecryptRecord(currentRecord, currentPasswordP);

				cryptoPadHandle = DmQueryRecord(dbCryptoPadRef, currentRecord);

				packedP = MemHandleLock(cryptoPadHandle);
				UnpackRecord(&cryptoRecord, packedP);
				
				/* Get the font settings */

				if (useSingleFont == separateFont) {
					tableFont = cryptoRecord.tableFont;
					fieldFont = cryptoRecord.fieldFont;
					tableColor = cryptoRecord.tableColor;
					fieldColor = cryptoRecord.fieldColor;
				}

				progrInfoP->statusFunc (progrInfoP,&progr,cryptoRecord.textP);

				/* Create a new record in MemoPad with the correct 
				 * text and change the structure of the CryptoPad record.
				 */

				ChangeRecordFromCryptoPadToMemoPad(cryptoRecord.textP, currentRecord, 
					tableFont, fieldFont, tableColor, fieldColor);
				
				MemHandleUnlock(cryptoPadHandle);
			}
		}
		currentRecord++;
	}

	progr.major.soFar = progr.major.total;
	progrInfoP->statusFunc (progrInfoP,&progr,NULL);

	RemoveCurrentPassword();

	sortOrder = GetMemoPadSortOrder(dbMemoPadRef);
	if (sortOrder == soAlphabetic) {
		DmQuickSort(dbCryptoPadRef, 
			(DmComparF*) &CompareCryptoPadRecords, (Int16) sortOrder);
		DmQuickSort(dbMemoPadRef, 
			(DmComparF*) &CompareMemoPadRecords, (Int16) sortOrder);
	}

	currentRecord = noRecordSelected;

	FrmDeleteForm(formP);
	FrmGotoForm(frmMainForm);
}

/*****************************************
 * ProgressFormInitAndDraw:
 * Init and draw the progress bar
 * for encryption/decryption of
 * the category. 
 *****************************************
 */
static ProgressInfoPtr ProgressFormInitAndDraw (FormPtr frmP)
{
	ProgressInfoPtr infoP = NULL;

	ErrNonFatalDisplayIf (!frmP, "nil form");


	// Allocate and initialize the Progress Info structure

	infoP = MemPtrNew (sizeof (ProgressInfoType));
	ErrNonFatalDisplayIf (!infoP, "alloc failed");
	if (!infoP)
		return (NULL);

	MemSet (infoP, sizeof (*infoP), 0);

	// Get bounds of the progress message and gauge objects once

	FrmGetObjectBounds (frmP, FrmGetObjectIndex (frmP, gadProgressMessage),
			&infoP->messageR);

	WinDrawGrayRectangleFrame(simpleFrame, &infoP->messageR);

	FrmGetObjectBounds (frmP, FrmGetObjectIndex (frmP, gadProgressBar),
					  &infoP->gaugeR);


	// Init the rest of the progress info structure

	infoP->progressForm = frmP;
	infoP->gaugeWidth = infoP->gaugeR.extent.x;
	infoP->gaugeLevel = 0;
	infoP->statusFunc = ProgressCallback;

	// Draw the form

	FrmDrawForm (frmP);

	// Initialize the gauge

	ProgressUpdateGauge (&infoP->gaugeR, 0);

	return (infoP);
}

/***************************************
 * ProgressCallback:
 * callback function for
 * the progress bar.
 ***************************************
 */
static Boolean ProgressCallback (void* progressInfoP, 
		ProgrType* progressP, Char* messageP)
{
	UInt16		gaugeLevel = 0;
	ProgressInfoPtr infoP = (ProgressInfoPtr) progressInfoP;
	UInt16		itemWidth = 0;


	// Process any pending events, such as those for our Stop button
 
	// Reset the auto-off timer to make sure we don't fall asleep in the
	//  middle of the update

	EvtResetAutoOffTimer ();

	while (ProgressHandleEvent (0 /*timeout*/) != nilEvent)
		{/*intentionally left blank*/}


	// Update the progress gauge

	if (progressP) {
		// Update the progress gauge counts
		ErrNonFatalDisplayIf (!progressP->major.total, "major denominator is zero");

		if (progressP->major.total)
			gaugeLevel = (UInt16) ((progressP->major.soFar * infoP->gaugeWidth)
								  / progressP->major.total);

		if (progressP->minor.soFar) {
			ErrNonFatalDisplayIf (!progressP->minor.total, "minor denominator is zero");

			if (progressP->major.total)
				itemWidth = (UInt16)(infoP->gaugeWidth / progressP->major.total);
			if (itemWidth && progressP->minor.total)
				gaugeLevel += (UInt16) ((progressP->minor.soFar * itemWidth)
									  / progressP->minor.total);
		}

		if (gaugeLevel != infoP->gaugeLevel) {
			infoP->gaugeLevel = gaugeLevel;
		  
		ProgressUpdateGauge (&infoP->gaugeR, gaugeLevel);
	}

} 


	// Update the progress message
	if (messageP) {
		FontID	currFont;

		// Switch to bold, saving the current font
		currFont = FntSetFont (stdFont);

		WinEraseRectangle (&infoP->messageR, 0 /*cornerDiam*/);

		DrawStringWithEllipsis (messageP, infoP->messageR.topLeft.x, 
			infoP->messageR.topLeft.y, infoP->messageR.extent.x, justCenter);

		// Restore font
		FntSetFont (currFont);
	}

	return (false);

} 

/**************************************
 * ProgressUpdateGauge:
 * Update the progress indicator.
 **************************************
 */
static void ProgressUpdateGauge (RectangleType*	gaugeRP, 
										UInt32 gaugeLevel)
{
	RectangleType 	eraseR;
	RectangleType 	completeR;

  
	ErrNonFatalDisplayIf (!gaugeRP, "null arg");


	// Draw the frame of gadget

	WinDrawRectangleFrame (simpleFrame, gaugeRP);
  

	// Fill in the gauge.

	if (gaugeLevel > (UInt32)gaugeRP->extent.x) {
		ErrDisplay ("gaugeLevel out of bounds");
		gaugeLevel = gaugeRP->extent.x;
	}

	completeR.topLeft.x = gaugeRP->topLeft.x;
	completeR.topLeft.y = gaugeRP->topLeft.y;
	completeR.extent.x = (UInt16)gaugeLevel;
	completeR.extent.y = gaugeRP->extent.y;
	WinDrawRectangle (&completeR, 0);
  

	// Erase the empty portion of the gauge (not really necessary).

	eraseR.topLeft.x = (UInt16)(gaugeRP->topLeft.x + gaugeLevel);
	eraseR.topLeft.y = gaugeRP->topLeft.y;
	eraseR.extent.x = (UInt16)(gaugeRP->extent.x - gaugeLevel);
	eraseR.extent.y = gaugeRP->extent.y;
	WinEraseRectangle (&eraseR, 0);

	return;
}


/******************************************
 * MainFormDisplayLookupField:
 * Show or hide the lookup field.
 ******************************************
 */

static void MainFormDisplayLookupField(void)
{
	FormPtr formP;
	UInt16 fieldIndex;
	UInt16 labelIndex;
	RectangleType rect;

	Coord xButton;
	Coord yButton;
	UInt16 buttonIndex;

	formP = FrmGetActiveForm();
	fieldIndex = FrmGetObjectIndex(formP, fldMainLookup);
	labelIndex = FrmGetObjectIndex(formP, lblMainLookup);

	buttonIndex = FrmGetObjectIndex(formP, btnMainNewMemo);
	FrmHideObject(formP, buttonIndex);
	FrmGetObjectPosition(formP, buttonIndex, &xButton, &yButton);

	// Display the lookup field or not.

	if (!useLookupField) {
		FrmHideObject(formP, fieldIndex);
		FrmHideObject(formP, labelIndex);

		// Set new Button Position

		xButton = 1;
		FrmSetObjectPosition(formP, buttonIndex, xButton, yButton);

	} else {
		FrmShowObject(formP, fieldIndex);
		FrmShowObject(formP, labelIndex);

		// Query Field bounds

		FrmGetObjectBounds(formP, fieldIndex, &rect);

		// Set new Button Position

		xButton = rect.topLeft.x + rect.extent.x + 5;
		FrmSetObjectPosition(formP, buttonIndex, xButton, yButton);

		FrmSetFocus(formP, fieldIndex);
	}

	FrmShowObject(formP, buttonIndex);
}

/***********************************************************************
 * MainFormLookupString:
 * Highlight the record which contins
 * the string in the lookup field.
 ***********************************************************************/

 static Boolean MainFormLookupString (EventPtr event)
{
	FormPtr	frm;
	UInt16 	fldIndex;
	FieldPtr fldP;
	Char * 	fldTextP;
  	TablePtr tableP;
	UInt16 	foundRecord = 0;
	Boolean completeMatch, match;
	Int16 	length;
   	Int16 	selectedRow, column;
   	RectangleType rec;
    UInt32 vers;
            
	frm = FrmGetActiveForm();
	fldIndex = FrmGetObjectIndex(frm, fldMainLookup);
	fldP = FrmGetObjectPtr (frm, fldIndex);

    // v4.11: Check for navigation support... if it exists, then we shouldn't
    // be setting focus here, and we should only continue processing if
    // field has focus.
    if (FtrGet(hsFtrCreator, hsFtrIDNavigationSupported, &vers) ==
            ftrErrNoSuchFeature) {
        FrmSetFocus(frm, fldIndex);
    }
    else {
        if (FrmGetObjectId(frm, FrmGetFocus(frm)) != fldMainLookup) {
            return(false);
        }
    }

	if (FldHandleEvent (fldP, event) || event->eType == fldChangedEvent ||
            event->eType == frmObjectFocusTakeEvent) {
		fldTextP = FldGetTextPtr(fldP);
		tableP = FrmGetObjectPtr (frm, FrmGetObjectIndex(frm, tblMainRecordList));

        // To make sure that we always end up in consistent state regarding
        // row selection, we always unhighlight (there is at least one corner
        // case where then new selection ends up on the same row but different
        // page... if we don't first unhighlight, the system gets confused).

        TblUnhighlightSelection(tableP);
        if (RecordLookupString(dbCryptoPadRef, fldTextP, 
			currentCategory, &foundRecord, &completeMatch)) {
            MainFormSelectRecord(foundRecord, tableP);
        }
      
	    if (!completeMatch) {
			// Delete the last character added.
			length = FldGetTextLength(fldP);
			FldDelete(fldP, length - 1, length);

			SndPlaySystemSound (sndError);
		}

		return true;
 	}

	// Event not handled
	return false;
   
}


/************************************************************
 * findStrInLine:
 * Find the string passed in the first line of the memo.
 *************************************************************/

Boolean findStrInLine (Char * targetP, Int16 targetLen, Char * strP, Int16 strLen)
{
  	Int16 len = StrLen ((char *) strP);
	Int16 count;

	if (!*strP) 
		return false;
  
	for (count = 0; *targetP && count < targetLen; count++) {
		if (*targetP == linefeedChr)
			return false;
    
	if ((*targetP & (~0x20)) == (*strP & (~0x20))) {
		if (StrNCaselessCompare (targetP, strP, len) == 0)
		return true;
	}    
		targetP++;
	}
	return false;
}

/************************************************************
 *  RecordLookupString:
 *  Search in each memo for the string passed
 *  from the lookup field.
 *************************************************************/

static Boolean RecordLookupString(DmOpenRef dbP, Char * key,
   UInt16 category, UInt16* recordP, Boolean *completeMatch)
{
	Int16	numOfRecords;
	MemHandle	cryptoRecordH, recordH;
	char *	textP;
	Boolean	match;

	match = *completeMatch = false;

	// If there isn 't a key to search with stop the with the first record.
	if (key == NULL || *key == '\0') {
		*completeMatch = true;
		return false;
	}

	numOfRecords = DmNumRecordsInCategory(dbP, currentCategory);
	if (numOfRecords == 0) {
    	return false;
    }

	while (!match) {
		cryptoRecordH = DmQueryNextInCategory (dbCryptoPadRef, recordP, currentCategory);
		if (!cryptoRecordH)
			break;

		if (isEncrypted(*recordP, dbCryptoPadRef)) {
			CryptoPackedRecordPtr packedP;
			CryptoRecordType record;

			recordH = DmGetRecord(dbCryptoPadRef, *recordP);
			packedP = MemHandleLock(recordH);
			UnpackRecord(&record, packedP);

			textP = record.textP; 
		} else {
			UInt16 index;

			/* Get the index of the MemoPad record */

			index = GetMemoPadRecordIndex(*recordP);

			recordH = DmQueryRecord(dbMemoPadRef, index);
			textP = MemHandleLock(recordH);
		}

      /* search string in data */
      match = *completeMatch = findStrInLine(textP, 56, key, StrLen(key));

		MemHandleUnlock(recordH);
		if (isEncrypted(*recordP, dbCryptoPadRef))
			DmReleaseRecord(dbCryptoPadRef, *recordP, false);

		if (!match) (*recordP)++;
	}

	return match;
}

/***********************************************************************
 * MainFormSelectRecord:
 * Highlight the record where
 * the string of the lookup field was found.
 ***********************************************************************/

static void MainFormSelectRecord (UInt16 recordNum, TablePtr tableP)
{
   UInt16 row, column;
   FormPtr frm;

	// Don't change anything if the same record is selected
	if (TblGetSelection(tableP, &row, &column) &&
		recordNum == TblGetRowID (tableP, row))
		return;

   // See if the record is displayed by one of the rows in the table
   // A while is used because if TblFindRowID fails we need to
   // call it again to find the row in the reloaded table.
	while (!TblFindRowID(tableP, recordNum, &row)) {
		currentRecord = recordNum;

		MainFormLoadRecords();
		TblRedrawTable(tableP);
	}
   
   // Select the item
   TblSelectItem (tableP, row, kRecordColumn);
   
   currentRecord = recordNum;
   currentRow = row;
} 
  
/***********************************************************************
 * MainFormClearLookupString:
 * Clear the lookup field.
 ***********************************************************************/

 static void MainFormClearLookupString (void)
{
	FormPtr frm;
	UInt16 fldIndex;
	FieldPtr fldP;
  	Int16 length;


	frm = FrmGetActiveForm();
	FrmSetFocus(frm, noFocus);
	fldIndex = FrmGetObjectIndex(frm, fldMainLookup);
	fldP = FrmGetObjectPtr (frm, fldIndex);

	length = FldGetTextLength(fldP);
	if (length > 0)
	FldDelete(fldP, 0, length);
}

/******************************************
 * MainFormGetNextCategory:
 * find the next category where there is
 * either a record from MemoPadDB or
 * from CryptoPad
 ******************************************
 */

static UInt16 MainFormGetNextCategory(UInt16 category)
{
	Boolean goToUnfiled = ((DmNumRecordsInCategory (dbCryptoPadRef, dmUnfiledCategory) != 0)
		&& (category != dmUnfiledCategory));

	if (currentRotateOrder != roAll) {
	/* The rotate order is either Categories-Unfiled
	 *  or All-Unfiled.
	 */

		if (CategoryGetNext(dbCryptoPadRef,  category) == dmAllCategories) {
			if (goToUnfiled) return dmUnfiledCategory;
		
			// The rotate order is Categories-Unfiled.

			if ((currentRotateOrder == roCategoriesUnfiled)
				&& !(goToUnfiled))
					return CategoryGetNext(dbCryptoPadRef,  dmAllCategories);
		}
	}

	return CategoryGetNext(dbCryptoPadRef, category);
}

/******************************************
 * MainFormNextCategory:
 * switches to the next category view
 ******************************************
 */

static void MainFormNextCategory(void)
{
	UInt16 category;

	category = MainFormGetNextCategory(currentCategory);

	if (category != currentCategory) {
		TablePtr tableP;
		ControlPtr controlP;
		FormPtr formP;

		if (category == dmAllCategories) showAllCategories = true;
		else showAllCategories = false;

		ChangeCategory(category);		

		/* Set the label of the category trigger */

		controlP = GetActiveFormObjectPtr(popMainCategory);
		CategoryGetName(dbMemoPadRef, currentCategory, categoryName);
		CategorySetTriggerLabel(controlP, categoryName);

		/* Display the new category */

		topRecord = 0;
		MainFormLoadTable ();
		tableP = GetActiveFormObjectPtr(tblMainRecordList);
		TblEraseTable (tableP);
		TblDrawTable (tableP);
	}
}

/******************************************
 * MainFormSelectCategory:
 * Invokes the standard category UI
 ******************************************
 */

static void MainFormSelectCategory(void)
{
	Boolean categoryUpdated;
	UInt16 category;
	FormPtr formP;

	// Process the category popup list. 

	category = currentCategory;
	
	formP = FrmGetActiveForm();
	categoryUpdated = CategorySelect(dbMemoPadRef, formP,
			popMainCategory, lstMainCategory, true,
			&category, categoryName, 1, categoryDefaultEditCategoryString);

	if (category == dmAllCategories) showAllCategories = true;
	else showAllCategories = false;

	if (categoryUpdated || (category != currentCategory)) {
		/* The category was changed, renamed or merged to
		 * another category.
		 */

		ChangeCategory(category);

		/* Clear the lookup field */
		MainFormClearLookupString();

		/* Display the new category */

        // Set a flag so 5-way nav code in event handler can keep category
        // pop-up selected instead of forcing table to be selected (consistent
        // with memopad behavior).
        didCategoryChange = true;

		FrmUpdateForm(frmMainForm, updateRedrawAll);	
	}
}

/******************************************
 * MainFormUpdateDisplay:
 * Updates the display of the main form.
 ******************************************
 */

static Boolean MainFormUpdateDisplay(UInt16 updateCode)
{	
	TablePtr tableP;

	if (updateCode & (updateTable | 
			updateFontChanged | updateRedrawAll)) {
		if (updateCode & updateTable)
			topRecord = 0;

		MainFormDisplayLookupField();

		MainFormLoadRecords();

		tableP = GetActiveFormObjectPtr(tblMainRecordList);
		TblEraseTable(tableP);
		TblDrawTable(tableP);		

		return (true);
	}
		
	return (false);
}

/******************************************
 * MainFormUpdateScrollBar:
 * Updates the size and position of the
 * scroll bar object.
 ******************************************
 */

static void MainFormUpdateScrollBar(void)
{
	FormPtr formP;
	TablePtr tableP;
	ScrollBarPtr scrollBarP;
	Int16 nRows;
	UInt16 nPosition = 0, maxValue;

	tableP = GetActiveFormObjectPtr(tblMainRecordList);
	scrollBarP = GetActiveFormObjectPtr(sclMainScrollBar);

	nRows = MainFormNumberOfRows(tableP);

	if (memosInCategory > nRows) {
		nPosition = DmPositionInCategory(dbCryptoPadRef, topRecord, currentCategory);
		maxValue = memosInCategory - nRows;		
	}
	else {
		maxValue = 0;
		nPosition = 0;
	}

	SclSetScrollBar(scrollBarP, nPosition, 0, maxValue, nRows);
}

/******************************************
 * MainFormScrollLines:
 * Scrolls the table nLines up (negative)
 * or down (positive).
 ******************************************
 */

static Boolean MainFormScrollLines(Int16 nLines)
{
	TablePtr tableP;
	Int16 recordNum;

    Boolean didMove = false;

	recordNum = topRecord;

	// Scroll down.
	if (nLines > 0)
		didMove = MainFormSeekRecord(&recordNum, nLines, dmSeekForward);
	// Scroll up. 
	else
		didMove = MainFormSeekRecord(&recordNum, -nLines, dmSeekBackward);

	topRecord = recordNum;
	currentRecord = noRecordSelected;

	tableP = GetActiveFormObjectPtr(tblMainRecordList);
	MainFormLoadRecords();

	TblRedrawTable(tableP);

    return(didMove);
}


/******************************************
 * MainFormScrollPage:
 * Scrolls the table one full page in
 * the given direction.
 ******************************************
 */

static Boolean MainFormScrollPage(WinDirectionType direction)
{
	TablePtr tableP;
	Int16 numRows;
	UInt16 recordNum;

	tableP = GetActiveFormObjectPtr(tblMainRecordList);
	numRows = MainFormNumberOfRows(tableP);

	recordNum = topRecord;
	currentRecord = noRecordSelected;

	// Scroll the table winDown a page (less one row).
	if (direction == winDown)
		{
		// Try going forward one page

		if (!MainFormSeekRecord (&recordNum, numRows - 1, dmSeekForward))
			{
			// Try going backwards one page from the last record

			recordNum = dmMaxRecordIndex;
			if (!MainFormSeekRecord (&recordNum, numRows - 1, dmSeekBackward))
				{
				// Not enough records to fill one page.  Start with the first record

				recordNum = 0;
				MainFormSeekRecord (&recordNum, 0, dmSeekForward);
				}
			}
		}
		
	// Scroll up a page (less one row).

	else
		{
		if (!MainFormSeekRecord (&recordNum, numRows - 1, dmSeekBackward))
			{
			// Not enough records to fill one page.  Start with the first record
			recordNum = 0;
			MainFormSeekRecord (&recordNum, 0, dmSeekForward);
			}
		}
	


	// Avoid redraw if no change

	if (topRecord != recordNum) {
		topRecord = recordNum;
		MainFormLoadRecords();
		TblRedrawTable(tableP);
        return(true);
	}
    return(false);
}


/******************************************
 * MainFormHandleMenuEvent:
 * Process menu commands.
 ******************************************
 */

static void MainFormHandleMenuEvent(UInt16 itemID)
{
	UInt32 currentRecordID;
	Boolean wasHiding;
	UInt16 mode;
	EventType event;
	FormPtr formP;

	switch (itemID) {
		case itemRecordEncryptCategory:
		case itemRecordDecryptCategory:
			if ((prefsGlobalPasswordP != 0) && (globalPasswordP == 0)) {
				event.eType = globalPasswordStatusEvent;
				event.data.generic.datum[1] = queryGlobalPassword;
				if (itemID == itemRecordEncryptCategory)
					event.data.generic.datum[2] = catDecrypted;
				else event.data.generic.datum[2] = catEncrypted;

				FrmPopupForm(frmQueryGlobalPasswordForm);
				EvtAddEventToQueue(&event);
			} else {
				event.eType = memoStatusEvent;
				if (itemID == itemRecordEncryptCategory)
					event.data.generic.datum[1] = catDecrypted;
				else event.data.generic.datum[1] = catEncrypted;

				FrmPopupForm(frmPasswordForm);
				EvtAddEventToQueue(&event);
			}
			break;

		case itemOptionsTableFont:
			event.eType = fontProcessEvent;
			event.data.generic.datum[1] = fontList;					
			FrmPopupForm(frmTableFontForm);
			EvtAddEventToQueue(&event);
			break;

		case itemRecordBeamCategory:
			MemoSendCategory(dbMemoPadRef, currentCategory);
			break;

		case itemOptionsPrefs:
			FrmPopupForm(frmPrefsForm);
			break;

		case itemOptionsSecurity:
			wasHiding = (PrivateRecordVisualStatus == hidePrivateRecords);
			 
			PrivateRecordVisualStatus = SecSelectViewStatus();
			
			if (wasHiding ^ (PrivateRecordVisualStatus == hidePrivateRecords)) 
			//xor on two logical values - mode to open DB has changed
				{
				// Close the application's data file.

				SetApplicationPrefs(0);
				DmCloseDatabase(dbMemoPadRef);
				DmCloseDatabase(dbCryptoPadRef);
			
				mode = (PrivateRecordVisualStatus == hidePrivateRecords) ?
					dmModeReadWrite : (dmModeReadWrite | dmModeShowSecret);
				
				dbCryptoPadRef = OpenOrCreateCryptoDatabase(mode);
				dbMemoPadRef = OpenMemoDatabase(mode);

				// Update the number of memo as some could have been hidden.
	
				if (showAllCategories)
					memosInCategory = DmNumRecordsInCategory (dbCryptoPadRef, dmAllCategories);
				else memosInCategory = DmNumRecordsInCategory (dbCryptoPadRef, currentCategory);

				// Read the preferences.

				GetApplicationPrefs(&currentRecordID);
			}
		
			//For safety, simply reset the currentRecord
			currentRecord = noRecordSelected;
			MainFormUpdateDisplay(updateTable);
			break;

		case itemOptionsAbout:
			formP = FrmInitForm(frmAboutForm);
			FrmDoDialog(formP);
			FrmDeleteForm(formP);
			break;

	}
}

/*************************************************
* ProgressHandleEvent:
* Get event from the progress
* bar form.
**************************************************
*/
static UInt32 ProgressHandleEvent (Int32 timeout)
{
	EventType event;
	UInt16 error;

	EvtGetEvent (&event, timeout);

	return (event.eType);
}

/*************************************************
* HighlightTable:
* v 4.11 code, add or remove highlights from
* main table as needed due to 5-way navigation
**************************************************
*/
static Boolean HighlightTable(FormPtr formP, Boolean addHighlight)
{
    UInt32 vers;
    RectangleType r;

    // Check for navigation support and to get version # (of course
    // we probably wouldn't be here if it weren't, but we need the
    // version # anyway
    if (FtrGet(hsFtrCreator, hsFtrIDNavigationSupported, &vers) != 0) {
        return(false);
    }
    // Have to check feature version to know which function
    // to use
    if (addHighlight) {
        FrmGetObjectBounds(formP, FrmGetObjectIndex(formP,
                    tblMainRecordList), &r);
        r.extent.x = 160;  // set to full width of form
        if (vers == 1) {
            HsNavDrawFocusRing(formP, tblMainRecordList,
                    hsNavFocusRingNoExtraInfo, &r,
                    hsNavFocusRingStyleHorizontalBars, false);
        }
        else {
            FrmNavDrawFocusRing(formP, tblMainRecordList,
                    hsNavFocusRingNoExtraInfo, &r,
                    hsNavFocusRingStyleHorizontalBars, false);
        }
    }
    else {
        if (vers == 1) {
            HsNavRemoveFocusRing(formP);
        }
        else {
            FrmNavRemoveFocusRing(formP);
        }
    }
    return(true);
}

/*************************************************
* HighlightObject:
* v 4.11 code, add highlights to catagory pop-up
* as needed due to 5-way navigation
**************************************************
*/
static Boolean HighlightObject(FormPtr formP, UInt16 object)
{
    UInt32 vers;
    RectangleType r;

    // Check for navigation support and to get version # (of course
    // we probably wouldn't be here if it weren't, but we need the
    // version # anyway
    if (FtrGet(hsFtrCreator, hsFtrIDNavigationSupported, &vers) != 0) {
        return(false);
    }

    FrmGetObjectBounds(formP, FrmGetObjectIndex(formP, object), &r);

    // Have to check feature version to know which function
    // to use
    if (vers == 1) {
        HsNavDrawFocusRing(formP, object,
                hsNavFocusRingNoExtraInfo, &r,
                hsNavFocusRingStyleObjectTypeDefault, false);
    }
    else {
        FrmNavDrawFocusRing(formP, object,
                hsNavFocusRingNoExtraInfo, &r,
                hsNavFocusRingStyleObjectTypeDefault, false);
    }
    return(true);
}

/*************************************************
* RowNotLast:
* v 4.11 code, given a row in the currently
* visible table, are there any usable records
* beyond it?
**************************************************
*/
static Boolean RowNotLast(TablePtr tableP, Int16 row)
{
	UInt16 recordNum;
    MemHandle record;

    recordNum = TblGetRowID(tableP, row);
    recordNum++;
    record = DmQueryNextInCategory(dbCryptoPadRef, &recordNum,
            currentCategory);
    if (record) {
        return(true);
    }
    return(false);
}

/*************************************************
* SetLookupSkip:
* v 4.11 code, have to dynamically set the "skip"
* state of the lookup field.
**************************************************

static void SetLookupSkip(FormPtr formP, Boolean no_skip)
{
    UInt16 afterObjectIDP;
    UInt16 aboveObjectIDP;
    UInt16 belowObjectIDP;
    FrmNavObjectFlagsType objectFlags;
    UInt32 vers;

    // Only do this if 5-way nav is available.
    if (FtrGet(hsFtrCreator, hsFtrIDNavigationSupported, &vers) != 0) {
        return;
    }

    FrmGetNavEntry(formP, fldMainLookup, &afterObjectIDP, &aboveObjectIDP,
            &belowObjectIDP, &objectFlags);
    if (no_skip) {
        objectFlags &= ~kFrmNavObjectFlagsSkip;
    }
    else {
        objectFlags |= kFrmNavObjectFlagsSkip;
    }
    FrmSetNavEntry(formP, fldMainLookup, afterObjectIDP, aboveObjectIDP,
            belowObjectIDP, objectFlags);
}
*/


Boolean MainFormHandleEvent(EventPtr eventP)
{
	Boolean handled = false;
	Boolean appHackBoolean;
	static Boolean wasShutDown = false;
	FormPtr	frm;
	TablePtr tableP;
	UInt16 row, column;
    Boolean tableHasFocus = false;
    Int16 lastRow;
    UInt32 vers;

	switch (eventP->eType) {
		/* A request to decrypt the record has
		 * been made. Swith to the Edit form
		 * to allow modification of the memo.
		 */
		case memoProcessEvent:
			ErrFatalDisplayIf(currentPasswordP == 0, "Password shouldn't be NULL");
			DecryptRecord(currentRecord, currentPasswordP);
			FrmGotoForm(frmEditForm);
			handled = true;
			break;

		case catProcessEvent:
			ErrFatalDisplayIf(currentPasswordP == 0, "Password shouldn't be NULL");
			if (eventP->data.generic.datum[1] == catDecrypted)
				EncryptCategory();
			else DecryptCategory();
			handled = true;
			break;

		/* We need to reload the records because a
		 * record has been deleted as a result of
		 * a forgotten password.
		 */

		case frmUpdateEvent:
			handled = MainFormUpdateDisplay (eventP->data.frmUpdate.updateCode);
            // Begin v4.11 5-way navigation code
            // Upon entering the main form, we need to set focus on table,
            // highlight it, and initialize interactive mode to false.
            if (handled) {
                tblInteractive = false;
                frm = FrmGetActiveForm();
                // SetLookupSkip(frm, useLookupField);
                if (didCategoryChange) {
                    tableP = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                                tblMainRecordList));
                    topRecord = TblGetRowID(tableP, TblGlueGetTopRow(tableP));
			        HighlightObject(frm, popMainCategory);
                    FrmSetFocus(frm, FrmGetObjectIndex(frm, popMainCategory));
                    didCategoryChange = false;
                }
                else {
			        HighlightTable(frm, true);
                    FrmSetFocus(frm, FrmGetObjectIndex(frm, tblMainRecordList));
                }
            }
            // End v4.11 5-way navigation code
			break;

		case ctlSelectEvent:
			switch (eventP->data.ctlSelect.controlID) {
				/* Creates a new memo and switch to the edit form */

				case btnMainNewMemo:
					if (CreateRecord())
						FrmGotoForm(frmEditForm);
					handled = true;
					break;

				/* Changes the currently selected category */

				case popMainCategory:
					MainFormSelectCategory();
					handled = true;
					break;
			}
			break;

        // Begin v4.11 5-way navigation code
        case frmObjectFocusTakeEvent:
            if (eventP->data.frmObjectFocusTake.objectID == tblMainRecordList) {
                frm = FrmGetFormPtr(eventP->data.frmObjectFocusTake.formID);
                tableP = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                            tblMainRecordList));
                if (! tblInteractive) {
			        HighlightTable(frm, true);
                    TblUnhighlightSelection(tableP);
                }
                else {
                    TblSelectItem(tableP, currentRow, kRecordColumn);
                }
                FrmSetFocus(frm, FrmGetObjectIndex(frm, tblMainRecordList));
                handled = true;
            }
            // We want the lookup field to reselect the correct table entry
            // if there is a search string active.
            if (eventP->data.frmObjectFocusTake.objectID == fldMainLookup) {
                frm = FrmGetFormPtr(eventP->data.frmObjectFocusTake.formID);
                FrmSetFocus(frm, FrmGetObjectIndex(frm, fldMainLookup));
                MainFormLookupString(eventP);
                handled = true;
            }
			break;

        // Because the system handling of the lookup field does not unhighlight
        // the table for some reason, we have to force the unhighlighting here.

        case frmObjectFocusLostEvent:
            if (eventP->data.frmObjectFocusLost.objectID == tblMainRecordList) {
                frm = FrmGetFormPtr(eventP->data.frmObjectFocusLost.formID);
                if ((FrmGetObjectId(frm, FrmGetFocus(frm)) == fldMainLookup) &&
                        (! tblInteractive)) {
                    HighlightTable(frm, false);
                }
            }
            break;

        // End v4.11 5-way navigation code

		/* A record has been tapped in the table 
		* or drag the memo with the stylus 
		*/

		case tblEnterEvent:
            // If 5-way Nav is supported and table is interactive mode, we
            // should highlight the selected record.
            tableP = eventP->data.tblSelect.pTable;
            if (tblInteractive) {
                TblUnhighlightSelection(tableP);
            }
            tblInteractive = true;
            currentRow = eventP->data.tblSelect.row;
			MainFormDragMemo(eventP);
            TblSelectItem(tableP, currentRow, kRecordColumn);
			handled = true;
			break;

		/* Scrolls the records */

		case sclRepeatEvent:
			MainFormScrollLines(eventP->data.sclRepeat.newValue -
						eventP->data.sclRepeat.value);
			/* Let the system repeat the event */
			break;
		
		
		case fldChangedEvent:
        	MainFormLookupString(eventP);
         	handled = true;
         	break;

		/* Handle the hard key button because of the bug of AppHack */

		case AppHackBugEvent:
			// MemoPad hard key button pressed?

			appHackBoolean = eventP->data.generic.datum[1];

			if (appHackBoolean) {
				if (!wasShutDown) {
					MainFormNextCategory ();

					/* Clear the lookup field */
					MainFormClearLookupString();
				} else wasShutDown = false;
			} else {
				/* The MemoPad hard key button have been
				 * pressed while the Palm was shut down
				 * so we don't need to change of category.
				 * A event have been sent and wasShutDown
				 * is set to true to stop the next AppHackBug
				 * event to prevent the category to change.
				 */

				wasShutDown = true;
			}
			handled = true;
			break;

		case keyDownEvent:
			frm = FrmGetActiveForm ();
			tableP = FrmGetObjectPtr (frm, FrmGetObjectIndex(frm,
                        tblMainRecordList));
            if (FrmGetObjectId(frm, FrmGetFocus(frm)) ==
                    tblMainRecordList) {
                tableHasFocus = true;
            }
			// Memo button pressed?
			/* AppHackBudEvent take care of it now. */

//			if (TxtCharIsHardKey(eventP->data.keyDown.modifiers, eventP->data.keyDown.chr)) {
//				if (!(eventP->data.keyDown.modifiers & poweredOnKeyMask))
//					MainFormNextCategory ();
//				handled = true;
//			}

//			else

			switch (eventP->data.keyDown.chr) {

				case linefeedChr:
				    if (TblGetSelection (tableP, &row, &column))
				        MainFormEnterTableItem();
				    handled = true;
				    break;
		
				// Scroll up key pressed?
				case vchrPageUp:
					MainFormScrollPage(winUp);
					handled = true;
					break;
	
				// Scroll down key pressed?
				case vchrPageDown:
					MainFormScrollPage(winDown);
					handled = true;
					break;
	
					// Send Data key pressed?
				case vchrSendData:
					MainFormHandleMenuEvent(itemRecordBeamCategory);
					handled = true;
					break;

                // Begin v4.11 5-way navigation code
				// Here we handle vchrRocker* events, which presumably only
                // happen when there is are 5-way nav buttons on the device.
                // We only need to handle rocker key events if the table has
                // focus.  The rest of the 5-way stuff is covered automatically
                // by the system software.

                case vchrRockerUp:
                    if (! tableHasFocus) break;
                    if (tblInteractive) {
                        // If scrolling actually moved us, then
                        // we handle the event, otherwise, we let
                        // it flow through and focus moves on.
                        TblUnhighlightSelection(tableP);
                        if (currentRow == TblGlueGetTopRow(tableP)) {
                            handled = MainFormScrollLines(-1);
                            if (handled) {
                                TblSelectItem(tableP, currentRow, kRecordColumn);
                            }
                        }
                        else {
                            currentRow--;
                            TblSelectItem(tableP, currentRow, kRecordColumn);
                            handled = true;
                        }
                    }
                    else {
                        // If scrolling actually moved us, then
                        // we handle the event, otherwise, we let
                        // it flow through and focus moves on.
                        // 
                        // There is a small as yet unsolved issue where
                        // we are at the top of the table already, but
                        // topRecord is not pointing to the first row.  In
                        // this case, the first push of the up button leaves
                        // in the same state.  At this point, the topRecord
                        // is at the right place, and next press does the right
                        // thing.  Letting this go for now as the amount of
                        // screwing around required to get it right may be
                        // substantial.
                        handled = MainFormScrollPage(winUp);
                    }
                    break;

                case vchrRockerDown:
                    if (! tableHasFocus) break;
                    if (tblInteractive) {
                        // Make sure we are not already at the end
                        // of the table.  If so, we let the system
                        // handle the event and focus moves on to the
                        // "new" button.
                        TblUnhighlightSelection(tableP);
                        if (RowNotLast(tableP, currentRow)) {
                            lastRow = TblGetLastUsableRow(tableP);
                            if ((currentRow < lastRow) &&
                                    (lastRow != tblUnusableRow)) {
                                currentRow++;
                                TblSelectItem(tableP, currentRow,
                                        kRecordColumn);
                            }
                            else {
                                MainFormScrollLines(1);
                                TblSelectItem(tableP, currentRow,
                                        kRecordColumn);
                            }
                            handled = true;
                        }
                    }
                    else {
                        // Make sure we are not already at the end
                        // of the table.  If so, we let the system
                        // handle the event and focus moves on to the
                        // "new" button or lookup field.
                        lastRow = TblGetLastUsableRow(tableP);
                        if (RowNotLast(tableP, lastRow)) {
                            MainFormScrollPage(winDown);
                            handled = true;
                        }
                    }
                    // We need special code here to handle the case where the
                    // lookup field is not present because the system nav stuff
                    // will only skip vertically past unusable objects, not
                    // horizontally.  An alternative solution would be to
                    // put the "new" button on a separate row below the
                    // lookup field, but then the up/down navigation isn't
                    // quite as intuitive when the lookup field is there.
                    // Note: adjusting the skip state of the lookup field
                    // object does not do anything to help this
                    // (e.g., SetLookupSkip).
                    if ((!handled) && (!useLookupField)) {
                        HighlightObject(frm, btnMainNewMemo);
                        FrmSetFocus(frm,
                                FrmGetObjectIndex(frm, btnMainNewMemo));
                        handled = true;
                    }
                    break;

                case vchrRockerCenter:
                    // If the lookup field is active and something is selected,
                    // enter that record instead of the system default, which
                    // is to toggle interactivity for the lookup field.
                    if (FrmGetObjectId(frm, FrmGetFocus(frm)) ==
                            fldMainLookup) {
                        if (TblGetSelection (tableP, &row, &column)) {
                            MainFormEnterTableItem();
				            handled = true;
                        }
                    }
                    if (! tableHasFocus) break;
                    if (tblInteractive) {
                        MainFormEditRecord(tableP, currentRow);
                    }
                    else {
                        tblInteractive = true;
                        HighlightTable(frm, false);
                        currentRow = TblGlueGetTopRow(tableP);
                        TblSelectItem(tableP, currentRow, kRecordColumn);
                    }
                    handled = true;
                    break;

                case vchrRockerLeft:
                    if (! tableHasFocus) break;
                    if (tblInteractive) {
                        tblInteractive = false;
                        HighlightTable(frm, true);
                        TblUnhighlightSelection(tableP);
                        handled = true;
                    }
                    break;

                case vchrRockerRight:
                    if (! tableHasFocus) break;
                    if (tblInteractive) {
                        TblUnhighlightSelection(tableP);
                    }
                    break;

//				default:
			}

            // v4.11: make sure the lookup field code does not get in the
            // way of processing the virtual key events used for 5-way nav.

            if (useLookupField && (! EvtKeydownIsVirtual(eventP))) {
                handled = MainFormLookupString(eventP);
            }

//				}

			// If printable character, create a new record.
			if ((!useLookupField) & (TxtCharIsPrint (eventP->data.keyDown.chr))) {
				if (CreateRecord ()) {
					WChar c;

					c =	eventP->data.keyDown.chr;
					if (c >= 'a' && c <= 'z')
						c -= ('a' - 'A');

					eventP->data.keyDown.chr = c;

					FrmGotoForm (frmEditForm);
					EvtAddEventToQueue (eventP);
				}
				handled = true;
			}
			break;

		case menuEvent:
			MainFormHandleMenuEvent(eventP->data.menu.itemID);
			return (true);
			break;

		case menuOpenEvent:
			if (OSVersion >= OS35) {
				MemHandle menuItemH;
				Char * menuItemP;

				menuItemH = DmGetResource('tSTR', strMenuSecurityItem);
				menuItemP = MemHandleLock(menuItemH);

				MenuAddItem (itemOptionsPrefs, itemOptionsSecurity, 'S', menuItemP);

				MemHandleUnlock(menuItemH);
				DmReleaseResource(menuItemH);
			}
			break;

		case menuCmdBarOpenEvent:
			MenuCmdBarAddButton(menuCmdBarOnLeft, BarSecureBitmap, menuCmdBarResultMenuItem, itemOptionsSecurity, 0);
			MenuCmdBarAddButton(menuCmdBarOnLeft, bmpBarFontBitmap, menuCmdBarResultMenuItem, itemOptionsTableFont, 0);

			// tell the field package to not add buttons automatically; we've done it all ourselves.
			eventP->data.menuCmdBarOpen.preventFieldButtons = true;
		
			// don't set handled to true; this event must fall through to the system.
			break;

		case frmOpenEvent:
			MainFormInit();
			FrmDrawForm(FrmGetActiveForm());
			handled = true;
			break;
	}

	return handled;
}
