#include "PrefsForm.h"
#include "Database.h"
#include "Utils.h"
#include "blowfish.h"

static UInt16 updateCode = 0;
static Boolean newUseSingleFont;
static Boolean newUseLookupField;

static void PrefsSortMemos(void);
static void PrefsFormInit(void);
static void PrefsSetGlobalPasswordStatus(void);
static void PrefsFormRotateOrder(void);
static void PrefsShowLines(void);
static void PrefsUseSingleFont(void);
static void PrefsUseLookupField(void);

static void PrefsGlobalPassword(void);
static void PrefsForgottenPassword(void);

/************************************
 * PrefsFormInit:
 * Initialize the Preferences form
 ************************************
 */

static void PrefsFormInit(void)
{
	FormPtr formP;
	ListPtr listSortP;
	ControlPtr popSortP;
	ListPtr listRotateP;
	ControlPtr popRotateP;
	ControlPtr controlShowLinesP;
	ControlPtr controlUseSingleFontP;
	ControlPtr controlMoveRecordOnTopP;
	ControlPtr controlUseLookupFieldP;
	UInt16 sortOrder;


	formP = FrmGetActiveForm();

	// Display the correct sort order.
	sortOrder = GetMemoPadSortOrder(dbMemoPadRef);

	popSortP = FrmGetObjectPtr(formP, FrmGetObjectIndex(formP, popPrefsSortOrder));
	listSortP = FrmGetObjectPtr(formP, FrmGetObjectIndex(formP, lstPrefsSortOrder));

	CtlSetLabel(popSortP, LstGetSelectionText(listSortP, sortOrder));
	LstSetSelection(listSortP, sortOrder);

	// Display the correct rotate order.
	popRotateP = FrmGetObjectPtr(formP, FrmGetObjectIndex(formP, popPrefsRotateOrder));
	listRotateP = FrmGetObjectPtr(formP, FrmGetObjectIndex(formP, lstPrefsRotateOrder));

	CtlSetLabel(popRotateP, LstGetSelectionText(listRotateP, currentRotateOrder));
	LstSetSelection(listRotateP, currentRotateOrder);

	// Display the correct show line setting.
	controlShowLinesP = GetActiveFormObjectPtr(chkPrefsShowLines);
	CtlSetValue(controlShowLinesP, showLines);

	// Display the correct use single font setting.
	controlUseSingleFontP = GetActiveFormObjectPtr(chkPrefsUseSingleFont);
	CtlSetValue(controlUseSingleFontP, useSingleFont);

	// Display the move record on top setting.
	controlMoveRecordOnTopP = GetActiveFormObjectPtr(chkPrefsMoveRecordOnTop);
	CtlSetValue(controlMoveRecordOnTopP, moveRecordOnTop);

	// Display the use lookup string setting.
	controlUseLookupFieldP = GetActiveFormObjectPtr(chkPrefsUseLookupField);
	CtlSetValue(controlUseLookupFieldP, useLookupField);

	PrefsSetGlobalPasswordStatus();

	newUseLookupField = useLookupField;
}

/************************************
 * PrefsSetGlobalPasswordStatus:
 * Change the status of the 
 * global password (assigned or not).
 ************************************
 */

static void PrefsSetGlobalPasswordStatus(void)
{
	ControlPtr controlP;
	Char * strGlobalPassword;
	MemHandle strGlobalPasswordH;

	if (prefsGlobalPasswordP == 0) {
		// Display the status of the global password as Unassigned.
		strGlobalPasswordH = DmGetResource('tSTR', strGlobalPasswordUnassigned);
		strGlobalPassword = MemHandleLock(strGlobalPasswordH);
		
		controlP = GetActiveFormObjectPtr(selPrefsGlobalPassword);
		CtlSetLabel(controlP, strGlobalPassword);

	   	MemHandleUnlock(strGlobalPasswordH);   
		DmReleaseResource(strGlobalPasswordH); 
	} else {
		// Display the status of the global password as Assigned
		strGlobalPasswordH = DmGetResource('tSTR', strGlobalPasswordAssigned);
		strGlobalPassword = MemHandleLock(strGlobalPasswordH);
		
		controlP = GetActiveFormObjectPtr(selPrefsGlobalPassword);
		CtlSetLabel(controlP, strGlobalPassword);

	   	MemHandleUnlock(strGlobalPasswordH);   
		DmReleaseResource(strGlobalPasswordH); 
	}
}

/*******************************************
 * PrefsSortMemos:
 * Sorts the memo alphabetically.
 * The PrefsSortMemos function uses the
 * CompareRecords comparison.
 *******************************************
 */

static void PrefsSortMemos(void)
{
	UInt8 sortOrder;
	FormPtr formP;

	formP = FrmGetActiveForm();
	sortOrder = LstGetSelection (FrmGetObjectPtr(formP, FrmGetObjectIndex(formP, lstPrefsSortOrder)));

	if (GetMemoPadSortOrder(dbMemoPadRef) != sortOrder) {
		
		// The sort order has been changed
		if (sortOrder == soAlphabetic) {
			if (FrmAlert (altAlphabeticSortAlert) == 1)
					return;
		}
	
	updateCode = updateTable;

	SetMemoPadSortOrder(dbMemoPadRef, sortOrder);

	}
}

/************************************
 * PrefsShowLines:
 * Sets the show lines attribute.
 ************************************
 */

static void PrefsShowLines(void)
{
	ControlPtr controlP;

	controlP = GetActiveFormObjectPtr(chkPrefsShowLines);
	showLines = CtlGetValue(controlP);
}

/***********************************
 * PrefsUseLookupField:
 * Sets the use lookup field attribut
 ***********************************
 */

static void PrefsUseLookupField(void)
{
	ControlPtr controlP;

	controlP = GetActiveFormObjectPtr(chkPrefsUseLookupField);
	useLookupField = CtlGetValue(controlP);

	if (newUseLookupField != useLookupField)
		updateCode = updateRedrawAll;
}

/***********************************
 * PrefsMoveRecordOnTop:
 * Sets the move record on top attribut
 ***********************************
 */

static void PrefsMoveRecordOnTop(void)
{
	ControlPtr controlP;
	Boolean moveRecord;

	controlP = GetActiveFormObjectPtr(chkPrefsMoveRecordOnTop);
	moveRecord = CtlGetValue(controlP);

	if (moveRecord && (GetMemoPadSortOrder(dbMemoPadRef) == soAlphabetic)) {
		FrmAlert (altMoveRecordOnTopImpossibleAlert);
		moveRecord = recNotMove;
	}
	
	moveRecordOnTop = moveRecord;
}

/************************************
 * PrefsUseSingleFont:
 * Sets the use single font attribute.
 ************************************
 */

static void PrefsUseSingleFont(void)
{
	ControlPtr controlP;
	Err err = 0;
	UInt16 recordNum = 0;
	MemHandle recordH;

	controlP = GetActiveFormObjectPtr(chkPrefsUseSingleFont);
	newUseSingleFont = CtlGetValue(controlP);

	if (newUseSingleFont != useSingleFont) {
		if (newUseSingleFont == singleFont) {
			if (FrmAlert (altUseSingleFontAlert) == 1)
				return;

			/* Set default values for the font and
			 * color of the table and field.
			 */
			useSingleFont = singleFont;
			currentFieldFont = stdFont;
			currentTableFont = stdFont;
			currentTableColor = blackColor;
			currentFieldColor = blackColor;


			/* Set the font settings to standard font
			 * for the table and the field font.
			 */
			while (true) {
					recordH = DmQueryNextInCategory (dbCryptoPadRef, &recordNum, dmAllCategories);
 					if (!recordH) break;
	 				SetRecordFont(recordNum);
	 				recordNum++;
			}
		} else useSingleFont = separateFont;

		updateCode = updateTable;
	}
}

/************************************
 * PrefsForgottenPassword:
 * Delete the global password
 * and all memos crypted with it.
 ************************************
 */

static void PrefsForgottenPassword(void)
{
	Int16 recordNum = 0;
	Err err;
	MemHandle handle;

	while (true) {

		handle = DmQueryNextInCategory (dbCryptoPadRef, &recordNum, dmAllCategories);

		if (!handle) break;

		if (isEncrypted(recordNum, dbCryptoPadRef)) {
			CryptoPackedRecordPtr packedP;
			CryptoRecordType record;
			Char * passwordP;
			UInt16 passwdLen;

			handle = DmQueryRecord(dbCryptoPadRef, recordNum);
			packedP = MemHandleLock(handle);
			UnpackRecord(&record, packedP);

			passwordP = MemPtrNew(record.passwdLen + 1);
			StrCopy(passwordP, record.passwordP);

			DmReleaseRecord(dbCryptoPadRef, recordNum, false);

			if (StrCompare(passwordP, prefsGlobalPasswordP) == 0) {
				DmRemoveRecord(dbCryptoPadRef, recordNum);
				updateCode = updateTable;
				recordNum--;
			}
		} 

		recordNum++;
	}	
	
	prefsGlobalPasswordP = 0;
	globalPasswordP = 0;
}


/************************************
 * PrefsGlobalPassword():
 * Sets the global password
 * attribute.
 ************************************
 */

static void PrefsGlobalPassword(void)
{
	EventType event;

		if (prefsGlobalPasswordP == 0) {
			event.eType = globalPasswordStatusEvent;
			event.data.generic.datum[1] = globalPasswordNotSet;
			event.data.generic.datum[2] = frmPrefsForm;

			FrmPopupForm(frmSetGlobalPasswordForm);
			EvtAddEventToQueue(&event);
		} else {
			event.eType = globalPasswordStatusEvent;
			event.data.generic.datum[1] = globalPasswordChange;
			event.data.generic.datum[2] = frmPrefsForm;

			FrmPopupForm(frmQueryGlobalPasswordForm);
			EvtAddEventToQueue(&event);
		}
}


/************************************
 * PrefsRotateOrderSelect:
 * Pops the selection list and select
 * the current rotate order.
 ************************************
 */

static void PrefsRotateOrder(void)
{
	FormPtr formP;
	Int8 rotateOrder;

	formP = FrmGetActiveForm();
	rotateOrder = LstGetSelection (FrmGetObjectPtr(formP, FrmGetObjectIndex(formP, lstPrefsRotateOrder)));

	currentRotateOrder = rotateOrder;
}


Boolean PrefsFormHandleEvent(EventPtr eventP)
{
	Boolean handled = false;

	switch (eventP->eType) {
		case ctlSelectEvent:
			switch (eventP->data.ctlSelect.controlID) {
				case btnPrefsOK:
					PrefsSortMemos();
					PrefsShowLines();
					PrefsUseSingleFont();
					PrefsMoveRecordOnTop();
					PrefsUseLookupField();
					PrefsRotateOrder();
					FrmReturnToForm(frmMainForm);
					if (updateCode)
						FrmUpdateForm(frmMainForm, updateTable);
					handled = true;
					break;

				case btnPrefsForgotten:
					if (prefsGlobalPasswordP != 0) {
						if (FrmAlert (altConfirmDeleteForgottenPasswordAlert) == 0) {
							PrefsForgottenPassword();
							PrefsSetGlobalPasswordStatus();
						} 
					} else FrmAlert (altNoGlobalPasswordToDeleteAlert);
					handled = true;
					break;

				case btnPrefsCancel:
					FrmReturnToForm(frmMainForm);
					handled = true;
					break;

				case selPrefsGlobalPassword:
					PrefsGlobalPassword();
					handled = true;
					break;
			}

		case frmCloseEvent:
			/* Let the system handle the event */
			break;

		case frmOpenEvent:
			PrefsFormInit();
			FrmDrawForm(FrmGetActiveForm());
			handled = true;
			break;
	}

	return handled;
}

